"""
目的：マルチプロセスにおいて seed を利かすための実験
目標：load_study 直後に sampler を渡せるか確認
"""

import os
# from multiprocessing import Process
import optuna

here, me = os.path.split(__file__)
os.chdir(here)


def objective(trial):
    x = trial.suggest_float('x', 0, 1)
    y = trial.suggest_float('y', 0, 1)
    return x**2 + y**2


def f(sampler):
    _study = optuna.load_study(
        study_name=f'{me.replace(".py", "")}',
        storage=f'sqlite:///{me.replace(".py", ".db")}',
        sampler=sampler,
    )
    _study.optimize(objective, n_trials=3)
    return _study


if __name__ == '__main__':

    # sampler setting
    sampler = optuna.samplers.TPESampler(seed=42)

    # delete previous study
    try:
        optuna.delete_study(
            study_name=f'{me.replace(".py", "")}',
            storage=f'sqlite:///{me.replace(".py", ".db")}',
        )
    except:
        pass

    # study setting
    optuna.create_study(
        sampler=sampler,
        study_name=f'{me.replace(".py", "")}',
        storage=f'sqlite:///{me.replace(".py", ".db")}',
        # load_if_exists=True,
    )

    # load and then optimization
    study = f(sampler)

    # result
    for t in study.trials:
        print(t.params)


# 1 回目
# {'x': 0.3745401188473625, 'y': 0.9507143064099162}
# {'x': 0.7319939418114051, 'y': 0.5986584841970366}
# {'x': 0.15601864044243652, 'y': 0.15599452033620265}

# 2 回目
# {'x': 0.3745401188473625, 'y': 0.9507143064099162}
# {'x': 0.7319939418114051, 'y': 0.5986584841970366}
# {'x': 0.15601864044243652, 'y': 0.15599452033620265}

# 結果
# seed は効く

# 考察
# load_study に sampler を指定できるので都合がいい。
# ただし Process に渡すと同じ値を吐くのだろう。
# args を保存して seed = seed + proc_id とするのがいいだろう




from time import sleep

import numpy as np
from scipy.signal import find_peaks
from tqdm import tqdm

from pythoncom import com_error

from PyFemtet.opt import FemtetOptuna

from win32com.client import Dispatch
from win32com.client import constants

# S パラメータ・共振周波数計算用クラス
class SParameterCalculator:
    
    def __init__(self):
        self.freq = []
        self.S = []
        self.interpolated_function = None
        self.resonance_frequency = None
        self.minimum_S = None

    def get_result_from_Femtet(self, Femtet):
        # 前準備
        Femtet.OpenCurrentResult(True)
        Gogh = Femtet.Gogh
        
        # 各モードに対して周波数と S(1,1) を取得する
        mode = 0
        freq_list = []
        dB_S_list = []
        for mode in tqdm(range(Gogh.Hertz.nMode), '結果取得中'):
            # Femtet 結果画面のモード設定
            Gogh.Hertz.Mode = mode
            sleep(0.1) # mode 変更が反映されるまでの猶予時間
            # 周波数の取得
            freq = Gogh.Hertz.GetFreq().Real
            # S パラメータの取得
            comp_S = Gogh.Hertz.GetSMatrix(0, 0)
            norm = np.linalg.norm((comp_S.Real, comp_S.Imag))
            dB_S = 20 * np.log10(norm)
            # 結果の取得
            freq_list.append(freq)
            dB_S_list.append(dB_S)
        self.freq = freq_list
        self.S = dB_S_list

    def calc_resonance_frequency(self):
        # S に対して最初のピークを与える freq を計算
        x = -np.array(self.S)
        peaks, _ = find_peaks(x, height=None, threshold=None, distance=None, prominence=0.5, width=None, wlen=None, rel_height=0.5, plateau_size=None)
        from PyFemtet.opt.core import SolveError
        if len(peaks)==0:
            raise SolveError('ピークが検出されませんでした。')
        self.resonance_frequency = self.freq[peaks[0]]
        self.minimum_S = self.S[peaks[0]]

    def get_resonance_frequency(self, Femtet):
        self.get_result_from_Femtet(Femtet)
        self.calc_resonance_frequency()
        return self.resonance_frequency * 1e-9
        

# サイズ拘束用関数
def anntena_is_smaller_than_substrate(Femtet):
    ant_r = Femtet.GetVariableValue('ant_r')
    Sx = Femtet.GetVariableValue('sx')
    return Sx/2 - ant_r # 1 以上になるようにする

def port_is_inside_anntena(Femtet):
    ant_r = Femtet.GetVariableValue('ant_r')
    xf = Femtet.GetVariableValue('xf')
    return ant_r - xf # 1 以上になるようにする
    
def size(Femtet):
    Sx = Femtet.GetVariableValue('sx')
    return Sx**2
    
if __name__=='__main__':
    # S パラメータ計算用クラス
    s = SParameterCalculator()
    
    # 最適化連携クラス
    FEMOpt = FemtetOptuna() # ここで起動している Femtet が紐づけされます
    
    # 最適化の設定
    FEMOpt.add_parameter('ant_r', 10, 5, 100, memo='円形アンテナの半径')
    FEMOpt.add_parameter('sx', 50, 5, 100, memo='基板のサイズ')
    FEMOpt.add_parameter('xf', 5, 1, 100, memo='給電ポートの偏心量')
    FEMOpt.add_constraint(anntena_is_smaller_than_substrate, lower_bound=1)
    FEMOpt.add_constraint(port_is_inside_anntena, lower_bound=1)
    FEMOpt.add_objective(s.get_resonance_frequency, '第一共振周波数（GHz）', direction=3.5)
    FEMOpt.add_objective(size, '基板サイズ（mm2）')
    
    FEMOpt.set_process_monitor()
    
    FEMOpt.main()
    
    import matplotlib.pyplot as plt
    plt.show()
    



import webbrowser

import dash
from dash import dcc, html
from dash.dependencies import Output, Input
import plotly.graph_objs as go
import plotly.express as px

import dash_bootstrap_components as dbc

import datetime

import numpy as np
import pandas as pd

COLORS = dict(
    Primary='#0d6efd', # 青
    Secondary='#6c757d', # 灰色
    Success='#198754', # 緑
    Danger='#dc3545', # 赤
    Warning='#ffc107', # 黄
    Info='#0dcaf0', # 水色
    Light='#f8f9fa', # 白っぽい灰色
    Dark='#212529', # 黒っぽい灰色
    )


def update_scatter_matrix(FEMOpt):
    # data
    data = FEMOpt.history.copy()
    # parameter_names = FEMOpt.get_history_columns('parameter')
    objective_names = FEMOpt.get_history_columns('objective')

    data['拘束'] = data['fit'].astype(str)
    data.loc[data['fit']==True, '拘束'] = 'OK'
    data.loc[data['fit']==False, '拘束'] = 'NG'

    data['好ましい'] = data['non_domi'].astype(str)
    data.loc[data['non_domi']==True, '好ましい'] = 'OK'
    data.loc[data['non_domi']==False, '好ましい'] = 'NG'
    
    fig = px.scatter_matrix(
        data,
        dimensions=objective_names,
        color="拘束",
        color_discrete_map=dict(OK=COLORS['Primary'], NG=COLORS['Secondary']),
        symbol='好ましい',
        symbol_map=dict(OK='circle', NG='circle-open')
        )
    
    fig.update_traces(
        diagonal_visible=False,
        showupperhalf=False,
        )
    
    fig.update_layout(
        title_text='目的ペアプロット',
        )    

    return fig    


def update_hypervolume(FEMOpt):
    data = FEMOpt.history
    
    # plot
    trace = go.Scatter(
        x=data['n_trial'],
        y=data['hypervolume'],
        mode='lines+markers',
        text=[dt.strftime('終了時刻：%Y/%m/%d %H:%M:%S') for dt in data['time']]
        )

    # figure layout
    layout = go.Layout(
        title_text="ハイパーボリューム",
        )
            
    # figure
    fig = go.Figure(
        data=trace,
        layout=layout,
        )
    
    return fig


class DashProcessMonitor:

    def start(self):
        self.app.run(debug=True, host='localhost', port=8080)
        # webbrowser.open("http://localhost:8080")

    
    def __init__(self, FEMOpt):

        # 引数の処理
        self.FEMOpt = FEMOpt
        
        # application の準備
        self.app = dash.Dash(__name__, external_stylesheets=[dbc.themes.BOOTSTRAP])


        #### component と layout の設定
        graph_layout = dbc.Container(
            dbc.Row(
                [
                    dbc.Col(dcc.Graph(id='hypervolume-graph'), md=6),
                    dbc.Col(dcc.Graph(id='scatter-matrix-graph'), md=6),
                    ]
                )
            )

        toggle_button = dbc.Button('Toggle auto-update', id='toggle-button', n_clicks=0)

        layout = html.Div(
            [
                html.H1("最適化の進捗状況"),
                dcc.Interval(
                    id='interval-component',
                    interval=2*1000,  # in milliseconds
                    max_intervals=0  # max number of intervals
                    ),
                graph_layout,
                html.Label(id='status-label'),
                toggle_button,
                ]
            )

        self.app.layout = layout


        # callback（分割したほうが並列計算できるらしい）

        # hypervolume
        @self.app.callback(
            Output('hypervolume-graph', 'figure'),
            Input('interval-component', 'n_intervals'))
        def update_hv(_):
            return update_hypervolume(self.FEMOpt)

        # scatter matrix
        @self.app.callback(
            Output('scatter-matrix-graph', 'figure'),
            Input('interval-component', 'n_intervals'))
        def update_sm(_):
            return update_scatter_matrix(self.FEMOpt)


        # フラグが立っていれば自動更新しない
        @self.app.callback(
            [Output('interval-component', 'max_intervals'),
            Output('status-label','children'),
            Output('toggle-button', 'disabled'),],
            [Input('interval-component', 'n_intervals'),])
        def stop_interval(_):
            if self.FEMOpt.should_finish():
                max_intervals = 0
                status_text = '最適化が終了しました。'
                toggle_button_disable = True
            else:
                max_intervals = -1
                status_text = '最適化が実行中です。'
                toggle_button_disable = False
            return max_intervals, status_text, toggle_button_disable

        # 自動更新の on / off
        @self.app.callback(
            [
                Output('interval-component', 'disabled'),
                Output('toggle-button', 'children'),
                ],
            [Input('toggle-button', 'n_clicks')])
        def toggle_interval(n):
            if n % 2 == 0:
                return False, 'グラフの自動更新をオフ'
            else:
                return True, 'グラフの自動更新をオン'
    
    
    
    
if __name__=='__main__':
    class DummyFEMOpt:
        def __init__(self):
            self.history = pd.DataFrame(
                dict(
                    n_trials=np.arange(10),
                    obj1=np.random.rand(10),
                    obj2=np.random.rand(10),
                    obj3=np.random.rand(10),
                    prm1=np.random.rand(10),
                    prm2=np.random.rand(10),
                    non_domi=np.random.rand(10)>0.5,
                    fit=np.random.rand(10)>0.5,
                    hypervolume=np.random.rand(10),
                    n_trial=np.arange(10),
                    time=[datetime.datetime.now()]*10,
                    )
                )
        def get_history_columns(self, kind):
            if kind=='parameter':
                return ['prm1', 'prm2']
            if kind=='objective':
                return ['obj1', 'obj2', 'obj3']
        
        def should_finish(self):
            return False
    
    FEMOpt = DummyFEMOpt()
    dpm = DashProcessMonitor(FEMOpt)
    dpm.start()


from PySide6.QtWidgets import QDialog, QVBoxLayout, QPushButton, QLabel, QApplication
from PySide6.QtCore import QTimer, Qt
from typing import Callable, Any
from multiprocessing import Value # 型ヒントのためのみ
import sys
import webbrowser


def open_browser():
    webbrowser.open('http://localhost:8080')


class SimplestDialog(QDialog):
    def __init__(
            self,
            FEMOpt,
            interrput_flag:Value,
            get_close_flag:Callable[[], bool] or None = None,
            fun_to_update:[Callable[[], Any]] or None = None,
            ):
        self.FEMOpt = FEMOpt
        
        # UI の設定
        super().__init__()
        self.setWindowTitle("最適化")
        self.setWindowFlags(self.windowFlags() & ~Qt.WindowCloseButtonHint)

        self.layout = QVBoxLayout()

        self.label = QLabel("経過時間：0秒")
        self.layout.addWidget(self.label)

        self.button = QPushButton("最適化を中断")
        self.button.clicked.connect(self.on_button_clicked)
        self.layout.addWidget(self.button)

        self.setLayout(self.layout)

        # 1 秒ごとに更新する関数
        self.timer = QTimer()
        self.timer.timeout.connect(self.update_label)
        if fun_to_update is not None:
            for fun in fun_to_update:
                self.timer.timeout.connect(fun)
        self.timer.start(1000)
        
        # 終了判定
        self.get_close_flag = get_close_flag

        # 経過時間
        self.seconds = 0

    def update_label(self):
        self.seconds += 1

        # 中断フラグが立ってなければ経過時間を表示
        if self.FEMOpt.shared_interruption_flag.value==0:
            self.label.setText(f"経過時間：{self.seconds}秒")

        # そうでなければ終了中ですと表示
        else:
            text = '''終了中です。
  現在の計算がすべて終了し、関連する非表示の
  プロセスがすべて終了するとこのダイアログが閉じます。'''
            self.label.setText(text)

        # 終了判定
        if self.get_close_flag is not None:
            if self.get_close_flag():
                self.close()


    def on_button_clicked(self):
        self.button.setEnabled(False)
        self.FEMOpt.shared_interruption_flag.value = 1

    def closeEvent(self, event):
        super().closeEvent(event)

if __name__ == "__main__":
    # app = QApplication(sys.argv)
    app = QApplication.instance()
    if app == None:
        app = QApplication([])
    dialog = SimplestDialog()
    dialog.show()
    app.exec()
    # del dialog
    # del app
    # import gc
    # gc.collect()



import os
import datetime
import functools
import warnings

import numpy as np
import pandas as pd

from .core import FemtetOptimizationCore

import optuna
# optuna.logging.disable_default_handler()
from optuna.study import MaxTrialsCallback
# from optuna.trial import TrialState
from optuna.exceptions import ExperimentalWarning
# optuna.logging.disable_default_handler()

from scipy.stats.qmc import LatinHypercube

from multiprocessing import Process, Value

import gc

warnings.filterwarnings('ignore', category=ExperimentalWarning)


def generate_LHS(bounds, seed=None)->np.ndarray:
    '''
    d 個の変数の bounds を貰い、N 個のサンプル点を得る。
    N は d を超える最小の素数 p に対し N = p**2
    '''
    d = len(bounds)

    sampler = LatinHypercube(
        d,
        scramble=False,
        strength=2,
        # optimization='lloyd',
        optimization='random-cd',
        seed=seed,
        )
    
    LIMIT = 100

    def is_prime(p):
        for i in range(2, p):
            if p%i==0:
                return False
        return True
    
    def get_prime(minimum):
        for p in range(minimum, LIMIT):
            if is_prime(p):
                return p
        
    n = get_prime(d+1)**2
    data = sampler.random(n) # [0,1)
    
    for i, (data_range, datum) in enumerate(zip(bounds, data.T)):
        minimum, maximum = data_range
        band = maximum - minimum
        converted_datum = datum * band + minimum
        data[:,i] = converted_datum
    
    return data # data.shape = (N, d)

# https://optuna.readthedocs.io/en/stable/index.html
class FemtetOptuna(FemtetOptimizationCore):

    def main(
            # 実装固有
            self,
            timeout=None,
            n_trials=None,
            study_name=None,
            use_init_LHS=True, # Latin Hypercube Sampling を初期値にする
            # 共通
            n_parallel=1,
            ):
        '''抽象クラスの main が実行される前に実装固有の設定を行う関数'''

        # 宣言
        self._constraints = []

        # 引数の処理
        self.use_init_LHS = use_init_LHS
        self.n_parallel = n_parallel
        
        #### study name のセットアップ
        # study_name（ファイル名）の設定
        if study_name is None:
            self.study_name = os.path.splitext(os.path.basename(self.history_path))[0] + ".db"
        # csv 保存パスと同じところに db ファイルを保存する
        self.study_db_path = os.path.abspath(
            os.path.join(
                os.path.dirname(self.history_path),
                self.study_name
                )
            )
        self.storage_name = f"sqlite:///{self.study_db_path}"

        #### timeout or n_trials の設定
        self._n_trials = n_trials
        # self.optimize_callbacks = []
        if n_trials is not None:
            self._n_trials = self._n_trials//self.n_parallel
            # self.optimize_callbacks.append(MaxTrialsCallback(n_trials)) # なぜか正常に機能しない
        self.timeout = timeout
        
        #### study の設定
        self._setup_study()

        # 本番処理に移る
        if n_parallel is None:
            super().main()
        else:
            super().main(n_parallel)


    def _setup_study(self):
        #### sampler の設定
        self.sampler_class = optuna.samplers.TPESampler
        self.sampler_kwargs = dict(constraints_func=self._constraint_function)


        #### study ファイルの作成
        study = optuna.create_study(
            study_name=self.study_name,
            storage=self.storage_name,
            load_if_exists=True,
            directions=['minimize']*len(self.objectives),
            )

        #### 初期値の設定
        params = self.get_parameter('dict')
        study.enqueue_trial(params, user_attrs={"memo": "initial"})
        
        #### LHS を初期値にする
        if self.use_init_LHS:
            df = self.get_parameter('df')
            names = []
            bounds = []
            for i, row in df.iterrows():
                names.append(row['name'])
                lb = row['lbound']
                ub = row['ubound']
                bounds.append([lb, ub])
            data = generate_LHS(bounds, seed=self.seed)
            for datum in data:
                d = {}
                for name, v in zip(names, datum):
                    d[name] = v
                study.enqueue_trial(d, user_attrs={"memo": "initial Latin Hypercube Sampling"})        
        

    def _main(self, process_idx=0):
        # setup_study が生成した study に接続
        seed = self.seed
        if seed is not None:
            seed += process_idx
        sampler = self.sampler_class(
            seed=seed,
            **self.sampler_kwargs,
        )
        study = optuna.load_study(
            study_name=self.study_name,
            storage=self.storage_name,
            sampler=sampler

        )

        # 最適化の実行
        study.optimize(
            self._objective_function,
            timeout=self.timeout,
            n_trials=self._n_trials,
            # callbacks = self.optimize_callbacks,
            )
        

    #### 目的関数
    def _objective_function(self, trial):

        # 中断指令の処理
        if self.shared_interruption_flag.value == 1:
            self.release_FEM()
            from .core import UserInterruption
            raise UserInterruption('ユーザーによって最適化が取り消されました。')


        #### 変数の設定
        x = []
        df = self.get_parameter('df')
        for i, row in df.iterrows():
            name = row['name']
            lb = row['lbound']
            ub = row['ubound']
            x.append(trial.suggest_float(name, lb, ub))
        x = np.array(x)

        #### user_attribute の設定
        # memo
        try:
            # あれば
            message = trial.user_attrs["memo"]
        except (AttributeError, KeyError):
            # なければ
            message = ''
        # 拘束のアトリビュート（Prune などの対策、計算できたら後で上書きする）
        dummy_data = tuple([1 for f in self._constraints])
        trial.set_user_attr("constraint", dummy_data)
        
        #### strict 拘束の計算
        # is_calcrated の直後まで変数をアップデートしてはいけないため
        # 一度 buff に現在の変数を控え、処理が終わったら戻す
        # 退避+更新
        buff = self._parameter.copy()
        self._parameter['value'] = x
        # restore の注意 / strict 拘束の処理から抜けうるところには
        # すべてに以下の処理を配置すること
        # self._parameter = buff
        # Femtet 関係の FEM システムであれば変数を更新
        from .core import Femtet, ModelError
        if isinstance(self.FEM, Femtet):
            try:
                # ModelError が起きうる
                self.FEM.update_model(self._parameter)
            except ModelError:
                raise optuna.TrialPruned()
        # strict 拘束の計算
        val_lb_ub_fn_list = [[cns.calc(), cns.lb, cns.ub, cns.f.__name__] for cns in self.constraints if cns.strict==True]
        for val, lb, ub, fn in val_lb_ub_fn_list:
            if lb is not None:
                if not (lb <= val):
                    print(f'拘束 {fn} が満たされませんでした。')
                    print('変数の組み合わせは以下の通りです。')
                    print(self._parameter)
                    self._parameter = buff # Femtet は戻す必要ない、以下同文
                    raise optuna.TrialPruned()
            if ub is not None:
                if not (val <= ub):
                    self._parameter = buff
                    print(f'拘束 {fn} が満たされませんでした。')
                    print('変数の組み合わせは以下の通りです。')
                    print(self._parameter)
                    raise optuna.TrialPruned()
        self._parameter = buff

        #### 解析実行
        # ModelError 等が起こりうる
        from .core import ModelError, MeshError, SolveError
        try:
            # parameters を更新し、解析、目的関数、全ての拘束の計算
            ojectiveValuesToMinimize = self.f(x, optimization_message=message)
        except (ModelError, MeshError, SolveError):
            raise optuna.TrialPruned()
        
        #### 拘束の計算
        _constraint_values = [] # 非正なら OK
        for i, cns in enumerate(self.constraints):
            lb, ub = cns.lb, cns.ub
            if lb is not None: # fun >= lb  <=>  lb - fun <= 0
                _constraint_values.append(lb - cns.calc())
            if ub is not None: # ub >= fun  <=>  fun - ub <= 0
                _constraint_values.append(cns.calc() - ub)
        trial.set_user_attr("constraint", _constraint_values)
        
        # 一応
        gc.collect()

        # 結果
        return tuple(ojectiveValuesToMinimize)


    #### 拘束関数(experimental)
    def _constraint_function(self, trial):
        return trial.user_attrs["constraint"]








# -*- coding: utf-8 -*-
import numpy as np
from PyFemtet.opt import FemtetScipy, FemtetOptuna
from PyFemtet.opt.core import NoFEM, ModelError, MeshError, SolveError

np.random.seed(3)

def parabora(FEMObj):
    x = FEMObj.get_parameter('values')
    if np.random.rand()>.5:
        return (x**2).sum()
    else:
        raise ModelError

def parabora2(FEMObj):
    x = FEMObj.get_parameter('values')
    if np.random.rand()>.5:
        return ((x-2)**2).sum()
    else:
        raise ModelError

if __name__=='__main__':

    FEMOpt = FemtetOptuna(FEMClass=NoFEM)

    FEMOpt.add_parameter('x', 1, 0, 2)
    FEMOpt.add_objective(parabora, args=(FEMOpt,))
    FEMOpt.add_objective(parabora2, args=(FEMOpt,))
    FEMOpt.main(n_trials=100)
    print(FEMOpt.history) # およそ 1/4 が計算されていれば OK
    


"""Test to set RANDOM_SEED.

Return True if:
    set specific cases
"""
import os
os.chdir(os.path.split(__file__)[0])
import pickle
from PyFemtet.opt import FemtetOptuna
from PyFemtet.opt.core import NoFEM


def parabola(FEMOpt)->float:
    """N-dimensional parabola function.

    Parameters
    ----------
    FEMOpt : FemtetOptimizationCore

    Returns
    -------
    float

    """
    x = FEMOpt.get_parameter('values')
    return (x**2).sum()


if __name__=='__main__':

    FEMOpt = FemtetOptuna(FEM=NoFEM())
    FEMOpt.set_random_seed(42)
    FEMOpt.add_parameter('x', 0, 0, 1)
    FEMOpt.add_parameter('y', 0, 0, 1)
    FEMOpt.add_parameter('z', 0, 0, 1)
    FEMOpt.add_objective(parabola, 'N次元放物線', args=(FEMOpt,))
    FEMOpt.main(n_trials=30)

    h = FEMOpt.history

    # with open('random_set_seed/random_seed_42.pkl', 'wb') as f:
    #     pickle.dump(h, f)

    with open('random_set_seed/random_seed_42.pkl', 'rb') as f:
        ref_h = pickle.load(f)

    parameter_names = FEMOpt.get_history_columns('parameter')
    ref_parameter_set = ref_h[parameter_names]
    dif_parameter_set = h[parameter_names]
    result = (ref_parameter_set == dif_parameter_set).all().all()
    
    print(result)

    




import optuna

def objective(trial):
    x = trial.suggest_float('x', 0, 1)
    y = trial.suggest_float('y', 0, 1)
    return x**2 + y**2

sampler = optuna.samplers.TPESampler(seed=42)

study = optuna.create_study(
    sampler=sampler,
)

study.enqueue_trial(
    dict(x=.5, y=.5),
)

study.optimize(objective, n_trials=10)

for trial in study.trials:
    print(trial.params)




"""
目的：マルチプロセスにおいて seed を利かすための実験
目標：load_study で seed が効くかの確認
"""

import os
# from multiprocessing import Process
import optuna

here, me = os.path.split(__file__)
os.chdir(here)


def objective(trial):
    x = trial.suggest_float('x', 0, 1)
    y = trial.suggest_float('y', 0, 1)
    return x**2 + y**2


def f():
    _study = optuna.load_study(
        study_name=f'{me.replace(".py", "")}',
        storage=f'sqlite:///{me.replace(".py", ".db")}'
    )
    _study.optimize(objective, n_trials=3)
    return _study


if __name__ == '__main__':

    # sampler setting
    sampler = optuna.samplers.TPESampler(seed=42)

    # delete previous study
    try:
        optuna.delete_study(
            study_name=f'{me.replace(".py", "")}',
            storage=f'sqlite:///{me.replace(".py", ".db")}',
        )
    except:
        pass

    # study setting
    optuna.create_study(
        sampler=sampler,
        study_name=f'{me.replace(".py", "")}',
        storage=f'sqlite:///{me.replace(".py", ".db")}',
        # load_if_exists=True,
    )

    # load and then optimization
    study = f()

    # result
    for t in study.trials:
        print(t.params)


# 1 回目
# {'x': 0.5538925912833553, 'y': 0.8971144707986053}
# {'x': 0.625757123038565, 'y': 0.9201323806970965}
# {'x': 0.797215571593796, 'y': 0.10090499190934443}

# 2 回目
# {'x': 0.11484858968716272, 'y': 0.9607568405523439}
# {'x': 0.942774590879533, 'y': 0.028013017988325783}
# {'x': 0.6834696189236454, 'y': 0.6219035541669511}

# 結果
# seed は効かない

# 考察
# seed はやはり load_study した段階で reseed_rng() が呼ばれるために効かなくなる





import time
from threading import Thread
from multiprocessing import Process, Manager
import psutil

import win32process
import win32gui
from win32com.client import Dispatch


from femtetutils import util


pp = False



def _get_hwnds(pid):
    def callback(hwnd, hwnds):
        if win32gui.IsWindowVisible(hwnd) and win32gui.IsWindowEnabled(hwnd):
            _, found_pid = win32process.GetWindowThreadProcessId(hwnd)
            if found_pid == pid:
                hwnds.append(hwnd)
        return True
    hwnds = []
    win32gui.EnumWindows(callback, hwnds)
    return hwnds

def _get_pid(hwnd):
    if hwnd>0:
        _, pid = win32process.GetWindowThreadProcessId(hwnd)
    else:
        pid = 0
    return pid

def _get_pids(process_name):
    pids = [p.info["pid"] for p in psutil.process_iter(attrs=["pid", "name"]) if p.info["name"] == process_name]
    return pids



def Dispatch_Femtet(timeout=30):
    '''
    Femtet の Dispatch を行う。

    Returns
    -------
    Femtet : pyIDispatch
    pid : int
        process id. if error, pid = 0.

    '''

    #### Dispatch 開始
    if pp: print('start Dispatch')
    Femtet = Dispatch('FemtetMacro.Femtet') # Dispatch は「予約」

    #### Dispatch が成功したことを保証するための処理
    start = time.time()
    # timeout 又は Femtet の hWnd が 0 超になるまで繰り返す
    while True:
        # hwnd を調べる
        hwnd = Femtet.hWnd
        
        # 0 超なら OK
        if hwnd>0:
            if pp: print('Dispatch finished successfully')
            break
        
        # そうでなければタイムアウトの判定
        duration = time.time() - start
        if pp: print('Dispatch duration', duration)
        if duration > timeout:
            if pp: print('Femtet が正常に実行されていません。')
            return None, 0

        time.sleep(1)

    # hwnd があれば pid が取得できることは保証されるから待たなくていい
    if pp: print('Dispatch finished')
    pid = _get_pid(hwnd)

    return Femtet, pid


def _f(pid, subprocess_id, shared_flags):

    # Dispatch して正常な hwnd が得られるまで待ち、pid を調べる
    if pp: print(f'from subp{subprocess_id}; let us start Dispatch...')
    Femtet, mypid = Dispatch_Femtet()
        
    # Dispatch が終了していることを通知
    shared_flags[subprocess_id] = True

    # pid が目的のものなら即刻開放する
    if mypid==pid:
        if pp: print(f'from subp{subprocess_id}; my pid is hit. i will release {pid}')
        return 0
    
    # pid が 0 なら何にも関与しないので即刻終了する
    elif mypid==0:
        if pp: print(f'from subp{subprocess_id}; failed to connect Femtet. i will finish.')
        return 0
    
    # そうでなければメインプロセスが Dispatch を終えるまで解放しない
    else:
        while True:
            if shared_flags[-1]:
                break
            time.sleep(1)
        if pp: print(f'from subp{subprocess_id}; main process seems to have been Dispatched. i will finish')
        return 1


def Dispatch_Femtet_with_new_process():
    '''
    Dispatch Femtet する。
    femtetutils を用いて新しいプロセスを立て、
    そのプロセスと繋ぐことを保証する。

    Parameters
    ----------
    pp : TYPE, optional
        print_progress, if True, show progress info. The default is False.

    Raises
    ------
    Exception
        何らかの失敗

    Returns
    -------
    Femtet : pyIDispatch
    pid : int
        process id. if error, pid = 0.

    '''

    # Femtet 起動
    succeed = util.execute_femtet()
    if not succeed:
        raise Exception('femtetutils を用いた Femtet の起動に失敗しました')

    # pid 取得
    pid = util.get_last_executed_femtet_process_id()
    if pid==0:
        raise Exception('起動された Femtet の認識に失敗しました')
    if pp: print('pid', pid)
    
    # ウィンドウが出てくるまで待つ
    timeout = 30
    start = time.time()
    while True:
        # ウィンドウがちゃんと出てきている
        hwnds = _get_hwnds(pid)
        if len(hwnds)>0:
            break
        # タイムアウトした
        duration = time.time() - start
        if duration > timeout:
            if pp: print('Femtet が正常に起動されませんでした。')
            return None, 0
    
    # 目的の pid を持つ Femtet を取得
    Femtet, mypid = Dispatch_Femtet_with_specific_pid(pid)

    print(f'pid of Dispatch object of Femtet:{mypid}')
    return Femtet, mypid


def Dispatch_Femtet_with_specific_pid(pid):
    #### 目的の pid の Femtet を取得
    # Femtet のプロセスをすべて列挙し、目的のプロセス以外を
    # サブプロセスで Dispatch してブロックする

    # 存在する Femtet プロセスの列挙
    pids = _get_pids('Femtet.exe')
    if pp: print('existing pids', pids)

    # 子プロセスの準備
    with Manager() as manager:
        # フラグの準備
        shared_flags = manager.list()
        for _ in range(len(pids)+1): # [-1]は自プロセス用のフラグ
            shared_flags.append(False)
        if pp: print('initial shared_flags', shared_flags[:])

        # プロセスの準備
        processes = []
        for subprocess_id in range(len(pids)):
            p = Process(
                target=_f,
                args=(
                    pid,
                    subprocess_id,
                    shared_flags
                    )
                )
            p.start()
            processes.append(p)

        # 子プロセスの Dispatch 完了を待つ
        while True:
            # 全てが True していれば次に
            if all(shared_flags[:-1]):
                if pp: print('finished: shared_flags', shared_flags[:])
                break
            if pp: print('shared_flags', shared_flags[:])
            time.sleep(1)

        # 子プロセスの Dispatch 完了を待って Dispatch
        Femtet, mypid = Dispatch_Femtet()
        if pp: print('finally Dispatched pid', mypid)

        # Dispatch 完了を通知
        shared_flags[-1] = True

        # サブプロセスすべての正常終了を待つ
        for p in processes:
            p.join()
        
        return Femtet, mypid


if __name__=='__main__':
    Femtet, pid = Dispatch_Femtet_with_new_process()
    start = time.time()
    Femtet.LoadProject(r'C:\Users\mm11592\Documents\myFiles2\working\1_PyFemtetOpt\PyFemtetOptDevelopment\PyFemtetPrj\tests\simple_femtet\simple.femprj', True)
    end = time.time()
    print(end-start)
    
    
    


import os
os.chdir(os.path.split(__file__)[0])
from PyFemtet.opt import FemtetOptuna

# from win32com.client import Dispatch
# Femtet = Dispatch('FemtetMacro.Femtet')
# Femtet.OpenCurrentResult(True)
def max_displacement(Femtet):
    dy = Femtet.Gogh.Galileo.GetMaxDisplacement_py()[1]
    return dy*1000

def volume(Femtet):
    _, v = Femtet.Gogh.CalcVolume_py([0])
    return v * 1e9

def bottom_area_1(Femtet):
    _, v = Femtet.Gogh.CalcArea_py([0])
    return v * 1e6

def bottom_area_2(Femtet, FEMOpt):
    param = FEMOpt.get_parameter()
    a = param['w'] * param['d']
    return a


if __name__=='__main__':
    FEMOpt = FemtetOptuna(femprj_path='simple_femtet/simple.femprj')
    
    # add_parameter
    FEMOpt.add_parameter('w', 10, 5, 20)
    FEMOpt.add_parameter('d', 10, 5, 20)
    FEMOpt.add_parameter('h', 100, 50, 200)
    
    # add_objective
    FEMOpt.add_objective(max_displacement, '変位(mm)')
    FEMOpt.add_objective(volume, '体積(mm3)')

    # add non-strict constraint
    FEMOpt.add_constraint(bottom_area_1, '底面積1(mm2)', 100, strict=False)

    # add strict constraint
    FEMOpt.add_constraint(bottom_area_2, '底面積2(mm2)', 99, args=FEMOpt)

    # overwrite constraint
    FEMOpt.add_constraint(bottom_area_2, '底面積2(mm2)', 50, args=FEMOpt)
    
    FEMOpt.main(n_trials=12, n_parallel=3)
    
    # import optuna
    # FEMOpt.

    
    




from __future__ import annotations
from typing import Any, Callable
from abc import ABC, abstractmethod
from functools import wraps

from time import sleep
import datetime

import numpy as np

from ._visualizationNameSpace import *


class SimpleProcessMonitor:
    def __init__(self, FEMOpt):
        # FEMOpt を貰う
        self.FEMOpt = FEMOpt
        self.N = len(self.FEMOpt.objectives)
        # サブプロットを定義
        self.fig, axes = plt.subplots(self.N, 1, sharex=True, figsize=(3, 2))
        self.fig.suptitle('シンプル目的関数モニター')
        # ax を作って ilne と一緒に保持する
        self.lines = []
        for i in range(self.N):
            ax = self.fig.axes[i]
            line, = ax.plot([], [], 'o-', zorder=2) # メインの履歴
            scat = ax.scatter([], [], s=50, marker='x', color='black', lw=3, label='FEM error', zorder=3) # エラーが生じているとき
            subLine = ax.axhline(y=0, color='red', linestyle='--', lw=1, zorder=1)
            self.lines.append([line, subLine, scat])
            # スクロールした時に direction を移動させる関数を登録するためにイベントピッカーを登録する
            ax.set_picker('draw_event')
        # 登録したイベントピッカーを接続する
        self.fig.canvas.mpl_connect('draw_event', self.adjustDirectionPosition)
        
        # TODO:中断（experimental）
        def lmd(*args, **kwargs):
            self.FEMOpt.shared_interruption_flag.value = 1
        self.fig.canvas.mpl_connect('close_event', lmd)
        
        self.fig.tight_layout()

    def update(self):
        objectives = self.FEMOpt.objectives
        xdata = self.FEMOpt.history['time']
        if len(xdata)==0:
            plt.pause(0.5)
            return
        for ax, (line, subLine, scat), objective in zip(self.fig.axes, self.lines, objectives):
            # データの描画
            ydata = self.FEMOpt.history[objective.name].values
            line.set_data(xdata, ydata)
            subLine.set_ydata(ydata[-1]) # ylim 再設定の時に邪魔になるから
            
            # lim 再設定
            ax.set_ylabel(objective.name)
            ax.xaxis.set_major_formatter(mdates.DateFormatter('%m/%d\n%H:%M'))
            ax.relim()
            ax.autoscale_view(True, True, True)
            # ターゲット線の描画(get_limするから後で)
            minimum, maximum = ax.get_ylim()
            if objective.direction=='maximize':
                y = maximum - (maximum - minimum) * 0.01
            elif objective.direction=='minimize':
                y = minimum + (maximum - minimum) * 0.01
            else: # 指定値と見做す
                y = objective.direction
            subLine.set_ydata(float(y))
        if self.fig.canvas.figure.stale:
            self.fig.canvas.draw_idle()
        self.fig.canvas.start_event_loop(1)
        
    def adjustDirectionPosition(self, _):
        objectives = self.FEMOpt.objectives
        for ax, lines, objective in zip(self.fig.axes, self.lines, objectives):
            # ターゲット線の描画(get_limするから後で)
            subLine = lines[1]
            minimum, maximum = ax.get_ylim()
            if objective.direction=='maximize':
                y = maximum - (maximum - minimum) * 0.01
            elif objective.direction=='minimize':
                y = minimum + (maximum - minimum) * 0.01
            else: # 指定値と見做す
                y = objective.direction
            subLine.set_ydata(float(y))


class HypervolumeMonitor(UpdatableFigure):

    @wraps(UpdatableFigure.__init__)
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        # label の設定
        # ！all, selected 以外は bool を持つ history の column とする
        # ！xlabel, ylabel も history の column とする
        self.labels['all'] = 'ハイパーボリューム'
        self.labels['selected'] = '選択中'
        # title の設定
        self.text_suptitle = self.fig.suptitle('ハイパーボリュームプロット')
        # 作図
        self.create()
        self.update()
        plt.pause(0.1)

    def _plot(self, ax):
        #### 空の scatter の実行
        # 基本。灰色。
        ax.plot([], [], label=self.labels['all'],
                   marker='o',
                   # markersize=5,
                   # markerfacecolor='gray',
                   # markeredgecolor='gray',
                   color='gray',
                   lw=1,
                   zorder=1, picker=True
                   )
        # 選択。色なし、黄枠線。何よりも手前にあるべき。
        ax.scatter([], [], label=self.labels['selected'],
                   facecolors='none',
                   edgecolors='yellow',
                   s=60, lw=2, zorder=4, picker=True
                   )
    
    def create(self):
        #### プロットを作る
        # 配置を作る
        gs = self.fig.add_gridspec(1, 1)
        ax = self.fig.add_subplot(gs[0, 0])
        self.axes = np.array([[ax]])
        # 作図
        self._plot(self.axes[0,0])
        # ラベルなどの追加
        self.axes[0,0].set(xlabel='n_trial', ylabel='hypervolume')
        #### legend の設定
        axleg = self.axes[0,0]
        loc = 'upper right'
        axleg.legend(
            *self.axes[0,0].get_legend_handles_labels(),
            loc=loc,
            )

    def on_pick(self, event):
        # ax の取得
        ax:Axes = event.artist.axes
        # pick されたデータの取得
        xdata, ydata = super()._get_xydata(event)
        # データからインデックスを取得
        idx = int(xdata[0])
        # picked_idx の更新
        self.picked_idx = idx
        # figure の更新
        self.update()
        plt.pause(0.1)    


class MultiobjectivePairPlot(UpdatableFigure):
    
    @wraps(UpdatableFigure.__init__)
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.labels['non_domi'] = '非劣解'
        self.labels['fit'] = '拘束範囲内'
        self.labels['all'] = 'その他'
        self.labels['selected'] = '選択中'
        self.text_suptitle = self.fig.suptitle('多目的ペアプロット')
        self.create()
        self.update()
        plt.pause(0.1)

    def _plot(self, ax):
        #### 空の scatter の実行
        # 非劣解。色なし、赤色。拘束違反でありうる。
        ax.scatter([], [], label=self.labels['non_domi'],
                   facecolors='none',
                   edgecolors='orangered',
                   s=20, lw=1, zorder=3, picker=True
                   )
        # 拘束範囲内。青色、枠線無し
        ax.scatter([], [], label=self.labels['fit'],
                   facecolors='royalblue',
                   edgecolors='none',
                   s=20, lw=0, zorder=2, picker=True
                   )
        # 基本。灰色、枠線無し。
        ax.scatter([], [], label=self.labels['all'],
                   facecolors='gray',
                   edgecolors='none',
                   s=20, lw=0, zorder=1, picker=True
                   )
        # 選択。色なし、黄枠線。何よりも手前にあるべき。
        ax.scatter([], [], label=self.labels['selected'],
                   facecolors='none',
                   edgecolors='yellow',
                   s=60, lw=2, zorder=4, picker=True
                   )
    
    def create(self):
        #### 下半分のペアプロットを実行
        # 配置を作る
        objective_names = self.FEMOpt.get_history_columns('objective')
        n = len(objective_names)
        gs = self.fig.add_gridspec(n-1, n-1)
        self.axes = []
        # 下半分にプロットを行っていく
        for r in range(n-1):
            self.axes.append([])
            for c in range(n-1):
                ax = self.fig.add_subplot(gs[r, c])
                self.axes[-1].append(ax)
                # 下半分ならプロット
                if r>=c:
                    # 作図
                    self._plot(ax)
                    # 一番下なら xlabel を表示、そうでなければ xticklabels を非表示
                    if r==n-2:
                        ax.set_xlabel(objective_names[c])
                    else:
                        [text.set_visible(False) for text in ax.xaxis.get_ticklabels()]
                    # 一番左なら ylabel を表示、そうでなければ yticklabels を非表示
                    if c==0:
                        ax.set_ylabel(objective_names[r+1])
                    else:
                        [text.set_visible(False) for text in ax.yaxis.get_ticklabels()]
                # そうでなければ非表示
                else:
                    ax.set_visible(False)
        self.axes = np.array(self.axes)

        #### 軸の同期設定
        # 同じ column で sharex
        for c in range(n-1):
            ax1 = self.axes[0,c]
            for r in range(1,n-1):
                ax2 = self.axes[r,c]
                ax2.sharex(ax1)
        # 同じ row で sharey
        for r in range(n-1):
            ax1 = self.axes[r,0]
            for c in range(0, n-1):
                ax2 = self.axes[r,c]
                ax2.sharey(ax1)

        #### legend の設定
        # 一番右上を legend 用にする
        axleg = self.axes[0,-1]
        axleg.set_visible(True)
        if id(axleg)==id(self.axes[0,0]):
            loc = 'upper right'
        else:
            loc = 'center'
            axleg.axis('off')
        axleg.legend(
            *self.axes[0,0].get_legend_handles_labels(),
            loc=loc,
            )
        
    def get_idx(self, x, y, xlabel, ylabel):
        x, y = x[0], y[0]
        # xdata, ydata ともに一致する idx を返す
        xdata = self.FEMOpt.history[xlabel]
        ydata = self.FEMOpt.history[ylabel]
        match_idx_x, = np.where(xdata==x)
        match_idx_y, = np.where(ydata==y)
        match_idx = np.intersect1d(match_idx_x, match_idx_y)
        if len(match_idx)==0:
            return -1
        else:
            return match_idx[0]

    def on_pick(self, event):
        # ax の取得
        ax:Axes = event.artist.axes
        # pick されたデータの取得
        xdata, ydata = super()._get_xydata(event)
        xlabel, ylabel = self._search_xylabels_in_shared_axes(ax)
        # データからインデックスを取得
        idx = self.get_idx(xdata, ydata, xlabel, ylabel)
        self.picked_idx = idx

        self.update()
        plt.pause(0.1)






# 目的：外から与えられた値に対する scatter matrix の live-update
# 目標：live-update


import dash
from dash import Dash, dcc, html, Input, Output, callback
import plotly

import numpy as np
import pandas as pd

app = Dash(__name__)
app.layout = html.Div(
    html.Div(
        [
            dcc.Graph(id='live-update-graph'),
            dcc.Interval(
                id='interval-component',
                interval=1*1000, # in milliseconds
                n_intervals=0
                )
            ]
        )
    )


# Multiple components can update everytime interval gets fired.
@callback(
    Output('live-update-graph', 'figure'),
    Input('interval-component', 'n_intervals')
    )
def update_graph_live(n):
    
    # data
    _data = np.random.rand(5,5)
    data = pd.DataFrame(_data, columns='A B C D E'.split())

    # Create the graph with subplots
    fig = plotly.tools.make_subplots(rows=2, cols=1)
    # fig['layout']['margin'] = {
    #     'l': 30, 'r': 10, 'b': 30, 't': 10
    # }
    # fig['layout']['legend'] = {'x': 0, 'y': 1, 'xanchor': 'left'}
    fig.append_trace({
        'x': data['A'],
        'y': data['B'],
        'name': 'Altitude',
        'mode': 'lines+markers',
        'type': 'scatter'
    }, 1, 1)
    fig.append_trace({
        'x': data['A'],
        'y': data['C'],
        'text': data['D'],
        'name': 'Longitude vs Latitude',
        'mode': 'lines+markers',
        'type': 'scatter'
    }, 2, 1)

    return fig


if __name__ == '__main__':
    app.run(debug=True, host='localhost', port=8080)



# 目的：外から与えられた値に対する scatter matrix の live-update
# 目標：クラス化した app での scatter matrix の live-update


import dash
from dash import Dash, dcc, html, Input, Output, callback
import plotly
import plotly.graph_objects as go

import numpy as np
import pandas as pd


class DashApp:
    
    def __init__(self):
        # application の設定
        self.app = Dash(__name__)
        self.app.layout = html.Div(
            html.Div(
                [
                    dcc.Graph(id='live-update-graph'),
                    dcc.Interval(
                        id='interval-component',
                        interval=3*1000, # in milliseconds
                        n_intervals=0
                        )
                    ]
                )
            )
                
        # callback の設定
        @callback(
            Output('live-update-graph', 'figure'),
            Input('interval-component', 'n_intervals')
            )
        def update(_):
            return self.update_graph_live()


    def update_graph_live(self):
        
        # data
        _data = np.random.rand(5,5)
        data = pd.DataFrame(_data, columns='A B C D E'.split())
    
        # pairplot
        trace=go.Splom(
            dimensions=[dict(label=c, values=data[c]) for c in data.columns],
            diagonal_visible=False,
            showupperhalf=False,
            )
        
        # figure layout
        layout = go.Layout(
            title_text="random",
            )
                
        # figure
        fig = go.Figure(
            data=trace,
            layout=layout,
            )
        
        return fig
    
    def start(self):
        self.app.run(debug=True, host='localhost', port=8080)

if __name__ == '__main__':
    pm = DashApp()
    pm.start()



# 目的：外から与えられた値に対する scatter matrix の live-update
# 目標：再利用可能なコードスタイル

from time import sleep

import dash
from dash import Dash, dcc, html, Input, Output, callback
import plotly
import plotly.graph_objects as go

import numpy as np
import pandas as pd

from threading import Thread


def update_scatter_matrix(calculator):
    # data
    data = calculator.df

    # pairplot
    trace=go.Splom(
        dimensions=[dict(label=c, values=data[c]) for c in data.columns],
        diagonal_visible=False,
        showupperhalf=False,
        )
    
    # figure layout
    layout = go.Layout(
        title_text="random",
        )
            
    # figure
    fig = go.Figure(
        data=trace,
        layout=layout,
        )
    
    return fig


def update_another_graph(calculator):
    ...
    return update_scatter_matrix(calculator)


class DashApp:

    def start(self):
        self.app.run(debug=True, host='localhost', port=8080)
    
    def __init__(self, calculator):
        
        self.calculator = calculator
        
        # application の設定
        self.app = Dash(__name__)

        # layout の設定
        self.app.layout = html.Div(
            html.Div(
                [
                    dcc.Graph(id='live-update-graph'),
                    dcc.Graph(id='live-update-graph2'),
                    dcc.Interval(
                        id='interval-component',
                        interval=1*1000, # in milliseconds
                        n_intervals=0
                        )
                    ]
                )
            )
                
        # callback の設定
        @callback(
            Output('live-update-graph', 'figure'),
            Output('live-update-graph2', 'figure'),
            Input('interval-component', 'n_intervals')
            )
        def update(_):
            ret = (
                update_scatter_matrix(self.calculator),
                update_another_graph(self.calculator)                
                )
            return ret


class Calculator:
    def calc(self):
        for i in range(10):
            _data = np.random.rand(5,5)
            self.df = pd.DataFrame(_data, columns='A B C D E'.split())
            sleep(1)


if __name__ == '__main__':
    c = Calculator()
    
    t = Thread(target=c.calc)
    t.start()
    
    pm = DashApp(c)
    pm.start()
    
    t.join()
    
    



import numpy as np

from PyFemtet.opt import FemtetOptuna
from PyFemtet.opt.core import NoFEM
from PyFemtet.opt.visualization import MultiobjectivePairPlot

def x(FEMOpt):
    r, theta, fai = FEMOpt.get_parameter('value')
    return r * np.cos(theta) * np.cos(fai)

def y(FEMOpt):
    r, theta, fai = FEMOpt.get_parameter('value')
    return r * np.cos(theta) * np.sin(fai)

def z(FEMOpt):
    r, theta, fai = FEMOpt.get_parameter('value')
    return r * np.sin(theta)

def strict_constraint(FEMOpt):
    r, theta, fai = FEMOpt.get_parameter('value')
    return fai

def weak_constraint(FEMOpt):
    return x(FEMOpt)

FEM = NoFEM()
FEMOpt = FemtetOptuna(FEM)

FEMOpt.add_parameter('r', 0.7, 0, 1)
FEMOpt.add_parameter('theta', 0, -np.pi/2, np.pi/2)
FEMOpt.add_parameter('fai', 0, 0, 2*np.pi)

FEMOpt.add_objective(x, 'x', args=FEMOpt)
FEMOpt.add_objective(y, 'y', args=FEMOpt)
FEMOpt.add_objective(z, 'z', args=FEMOpt)

FEMOpt.add_constraint(strict_constraint, 'fai<pi', upper_bound=np.pi, args=FEMOpt)
FEMOpt.add_constraint(weak_constraint, 'x>0', lower_bound=0, strict=False, args=FEMOpt)

FEMOpt.set_process_monitor(MultiobjectivePairPlot)

FEMOpt.main(n_trials=100)




import numpy as np

from PyFemtet.opt.visualization import MultiobjectivePairPlot
from PyFemtet.opt.visualization import UpdatableSuperFigure
from PyFemtet.opt.visualization import HypervolumeMonitor

from PyFemtet.opt import FemtetOptuna
from PyFemtet.opt.core import NoFEM


if __name__=='__main__':
    np.random.seed(4)
    
    FEM = NoFEM()
    FEMOpt = FemtetOptuna(FEM)
    
    n = 2
    
    for i in range(n):
        FEMOpt.add_objective(lambda:None, f"obj{i}")
    
    FEMOpt._initRecord()
    
    for i in range(n):
        FEMOpt.history[f'obj{i}'] = np.random.rand(10)*np.random.rand()*10
    
    FEMOpt._calcNonDomi()
    FEMOpt._calc_hypervolume()
    FEMOpt.history['fit'] = True
    idx = (np.random.rand(1)*10).astype(int)
    FEMOpt.history['fit'][idx] = False
    FEMOpt.history['n_trial'] = range(10)
    
    # MultiobjectivePairPlot(FEMOpt)
    # HypervolumeMonitor(FEMOpt)
    pm = UpdatableSuperFigure(FEMOpt, HypervolumeMonitor, MultiobjectivePairPlot)
    


from PyFemtet.opt import FemtetOptuna
from PyFemtet.opt.visualization import HypervolumeMonitor

from win32com.client import Dispatch, constants


def fundamental_resonance(Femtet):
    '''mode[0] の共振周波数の取得'''
    # 結果取得用関数の作成に使用した前処理
    # Femtet = Dispatch('FemtetMacro.Femtet')
    # Femtet.OpenCurrentResult(True)
    Gogh = Femtet.Gogh
    Gogh.Galileo.Mode = 0
    freq:'complex' = Gogh.Galileo.GetFreq()
    # print(freq.Real)
    return freq.Real

def mass(Femtet):
    rho = Femtet.GetVariableValue('rho')
    _, volume = Femtet.Gogh.CalcVolume_py([0])
    return rho * volume # kg/m3 * m3

def thickness(Femtet):
    external_r = Femtet.GetVariableValue('external_r')
    internal_r = Femtet.GetVariableValue('internal_r')
    # Femtet.Gogh
    return external_r - internal_r
    

if __name__=='__main__':

    FEMOpt = FemtetOptuna()

    FEMOpt.add_parameter('internal_r', initial_value=1.1, lower_bound=0.2, upper_bound=3)
    FEMOpt.add_parameter('external_r', initial_value=1.5, lower_bound=0.1, upper_bound=2.9)
    FEMOpt.add_parameter('rho', initial_value=8000, lower_bound=6000, upper_bound=9000)

    FEMOpt.add_objective(fundamental_resonance, '共振周波数(Hz)', 50)
    FEMOpt.add_objective(mass, '質量(kg)', 1000)
    
    FEMOpt.add_constraint(thickness, '厚さ(m)', lower_bound=0.3)
    
    FEMOpt.set_process_monitor()
    
    FEMOpt.main()
    
    # import matplotlib.pyplot as plt
    # plt.show() # メインルーチンをここで止め、プロセスモニタが消えないようにする
    








import numpy as np
import pandas as pd

# plot ライブラリ
import matplotlib.pyplot as plt
import matplotlib.gridspec as gridspec

# 日付
import matplotlib.dates as mdates

# dpi
import ctypes

scaleFactor = ctypes.windll.shcore.GetScaleFactorForDevice(0)

# フォント設定
from matplotlib import rcParams

rcParams["font.size"] = 9
rcParams["figure.dpi"] = scaleFactor
# rcParams['font.family'] = 'MS Gothic'

# スタイル
plt.style.use("ggplot")

# 3D
from mpl_toolkits.mplot3d import Axes3D

# hint
from abc import ABC, abstractmethod
from matplotlib.figure import Figure, SubFigure
from matplotlib.axes import Axes
from matplotlib.lines import Line2D  # plot
from matplotlib.collections import PathCollection  # scatter
from matplotlib.text import Text  # title, suptitle, etc
# 日本語
from japanize_matplotlib import japanize

japanize()

# FEMOpt class
from ..core import FemtetOptimizationCore

default_supfigsize = (7, 3)  # size(px) = figsize*dpi
default_figsize = (5, 4)


class ConcreteClassOfUpdatableFigure:
    '''
    UpdatableFigureの具象クラス。
    インスタンスではなくクラスであることに注意してください。
    '''
    pass


class UpdatableSuperFigure:

    def __init__(
            self,
            FEMOpt: FemtetOptimizationCore,
            UpdatableFigure1: 'ConcreteClassOfUpdatableFigure',
            UpdatableFigure2: 'ConcreteClassOfUpdatableFigure',
    ):

        self.FEMOpt = FEMOpt
        self.monitors = []

        # figure の設定
        self.fig = plt.figure(figsize=default_supfigsize)
        self.fig.canvas.mpl_connect('button_press_event', self.force_redraw)
        self.fig.suptitle('画面をクリックして表示を更新；要素の選択はできません')

        # subfigure の配置
        gs = self.fig.add_gridspec(1, 2)
        # ひだり
        sub_fig1 = self.fig.add_subfigure(gs[0, 0])
        monitor1 = UpdatableFigure1(self.FEMOpt, sub_fig1)
        self.monitors.append(monitor1)
        # みぎ
        sub_fig2 = self.fig.add_subfigure(gs[0, 1])
        monitor2 = UpdatableFigure2(self.FEMOpt, sub_fig2)
        self.monitors.append(monitor2)

        # 閉じたときの処理はどれかに対して行えばいい、あるいはクラスメソッドにするか
        self.fig.canvas.mpl_connect('close_event', monitor1.on_close)

        # ※ matplotlib 3.8.2 では subfigure の中の artist に対して pick_event が効かない
        # 修正はすでにされており 23.11.18 にテストが終了していたので、
        # 近々 main ブランチにマージされるのではないか。
        # for m in self.monitors:
        #     self.fig.canvas.mpl_connect('pick_event', m.on_pick)

    def update(self):
        for m in self.monitors:
            m.update()
        if self.fig.canvas.figure.stale:
            self.fig.canvas.draw_idle()
        # self.fig.canvas.start_event_loop(1)

    def force_redraw(self, *args, **kwargs):
        self.fig.canvas.draw()  # 再描画されないことがある？
        plt.pause(1)


class UpdatableFigure(ABC):

    def __init__(
            self,
            FEMOpt: FemtetOptimizationCore,
            fig: SubFigure = None
    ):

        self.FEMOpt = FEMOpt
        self.labels = {}
        self.picked_idx = -1

        if fig is None:
            self.fig = plt.figure(figsize=default_figsize)
            self.fig.canvas.mpl_connect('pick_event', self.on_pick)
            self.fig.canvas.mpl_connect('close_event', self.on_close)
        else:
            # as subfigure
            self.fig = fig

    def _update_range(self, xrange, xdata):
        if len(xdata) > 0:
            if xrange[0] is None:
                xrange = (min(xdata), max(xdata))
            else:
                xrange = (min(xrange[0], min(xdata)), max(xrange[1], max(xdata)))
        return xrange

    def _finilize_range(self, xlim, xrange, xdata, mergin=(0.1, 0.1)):
        '''
        mergin つき range と lim を比較し、範囲の広いほうを採用する
        '''
        if xrange[0] is None:
            return xlim

        width = xrange[1] - xrange[0]
        minimum = xrange[0] - width * mergin[0]
        maximum = xrange[1] + width * mergin[1]

        # データが 2 点だけならば桁を加味して強制的に xrange を採用する
        if len(xdata) == 2:
            return (minimum, maximum)

        minimum = min(minimum, xlim[0])
        maximum = max(maximum, xlim[1])

        return (minimum, maximum)

    def _search_xylabels_in_shared_axes(self, ax):
        '''ax に label がなければ share された ax から label を探す'''
        xlabel, ylabel = ax.get_xlabel(), ax.get_ylabel()

        axes = ax.get_shared_x_axes().get_siblings(ax)
        for shared_ax in axes:
            if xlabel == '':
                xlabel = shared_ax.get_xlabel()

        axes = ax.get_shared_y_axes().get_siblings(ax)
        for shared_ax in axes:
            if ylabel == '':
                ylabel = shared_ax.get_ylabel()

        return xlabel, ylabel

    def _get_labeled_artists(self, ax, label):
        artists = ax.get_children()
        filterd_artists = []
        for artist in artists:
            if hasattr(artist, 'get_label') == True:
                if type(artist) == PathCollection:
                    if artist.get_label() == label:
                        filterd_artists.append(artist)
                elif type(artist) == Line2D:
                    if artist.get_label() == label:
                        filterd_artists.append(artist)
        return filterd_artists

    def update(self):
        for ax in self.fig.axes:
            xlabel, ylabel = self._search_xylabels_in_shared_axes(ax)

            # data から計算する lim の初期化
            xrange, yrange = (None, None), (None, None)
            for key, label in self.labels.items():
                # picked がなければ selected は飛ばす
                if (key == 'selected') and (self.picked_idx == -1):
                    continue
                # label された artist 全てを plot
                labeld_artists = self._get_labeled_artists(ax, label)
                for artist in labeld_artists:
                    # all, selected は予約
                    if key == 'all':
                        idx = slice(None)
                    elif key == 'selected':
                        idx = self.picked_idx
                    else:
                        idx = self.FEMOpt.history[key] == True
                    xdata, ydata = self.FEMOpt.history[[xlabel, ylabel]].dropna().values[idx].T
                    # もし長さが 1 だと ndarray にならないので変換
                    if isinstance(xdata, float) or isinstance(xdata, int):  # np.float などでも可
                        xdata = np.array([xdata])
                    if isinstance(ydata, float) or isinstance(ydata, int):  # np.float などでも可
                        ydata = np.array([ydata])
                    # plot の更新
                    if isinstance(artist, Line2D):
                        artist.set_data(xdata, ydata)
                        xrange = self._update_range(xrange, xdata)
                        yrange = self._update_range(yrange, ydata)
                    elif isinstance(artist, PathCollection):
                        artist.set_offsets(np.array([xdata, ydata]).T)
                        xrange = self._update_range(xrange, xdata)
                        yrange = self._update_range(yrange, ydata)
            # 全てのラベルのプロットが終われば set_lim
            xrange = self._finilize_range(ax.get_xlim(), xrange, xdata)
            yrange = self._finilize_range(ax.get_ylim(), yrange, ydata)
            ax.set(
                xlim=xrange,
                ylim=yrange,
            )
        if type(self.fig) == Figure:
            self.text_suptitle.set_y(0.96)
            # self.fig.subplots_adjust(
            #     left=0.2,
            #     bottom=0.2,
            #     right=0.95,
            #     top=0.95,
            #     wspace=0.05,
            #     hspace=0.05
            #     )
            self.fig.tight_layout()
            if self.fig.canvas.figure.stale:
                self.fig.canvas.draw_idle()
            # self.fig.canvas.start_event_loop(1) # pause すると画面がそのたびアクティブになって邪魔
        elif type(self.fig) == SubFigure:
            self.text_suptitle.set_y(0.85)
            self.fig.subplots_adjust(
                left=0.2,
                bottom=0.2,
                right=0.95,
                top=0.8,
                wspace=0.05,
                hspace=0.05
            )

    @abstractmethod
    def create(self):
        pass

    def _get_xydata(self, event) -> [np.ndarray]:
        fig: Figure = event.canvas.figure
        ax: Axes = event.artist.axes
        artist = event.artist
        idx: np.ndarray = event.ind  # 選択したデータのインデックス
        if isinstance(artist, Line2D):
            xdata, ydata = [data[idx] for data in artist.get_data()]
        if isinstance(artist, PathCollection):
            xdata, ydata = artist.get_offsets()[idx].T
        # print(xdata, ydata)
        return xdata, ydata

    def on_close(self, *args, **kwargs):
        self.FEMOpt.shared_interruption_flag.value = 1






from threading import Thread

import webbrowser

import dash
from dash import dcc, html
from dash.dependencies import Output, Input
import plotly.graph_objs as go
import plotly.express as px

import datetime
from time import sleep

import numpy as np
import pandas as pd


def update_scatter_matrix(FEMOpt):
    # data
    data = FEMOpt.history
    parameter_names = FEMOpt.get_history_columns('parameter')
    objective_names = FEMOpt.get_history_columns('objective')
    # is_fit = data['fit']
    # is_nondomi = data['non_domi']

    # pairplot
    # text_list = []
    # for i, row in data[parameter_names].iterrows():
    #     text = ''
    #     for prm in row.index:
    #         value = row[prm]
    #         text = text + f'{prm} : {value:.3f}<br>'
    #     text_list.append(text)
    trace=go.Splom(
        dimensions=[dict(label=c, values=data[c]) for c in objective_names],
        diagonal_visible=False,
        showupperhalf=False,
        # text=text_list,
        )
    
    # figure layout
    layout = go.Layout(
        title_text="多目的ペアプロット",
        )
            
    # figure
    fig = go.Figure(
        data=trace,
        layout=layout,
        )
    
    return fig    


def update_hypervolume(FEMOpt):
    data = FEMOpt.history
    
    # plot
    trace = go.Scatter(
        x=data['n_trial'],
        y=data['hypervolume'],
        mode='lines+markers',
        # text=[dt.strftime('終了時刻：%Y/%m/%d %H:%M:%S') for dt in data['time']]
        )

    # figure layout
    layout = go.Layout(
        title_text="ハイパーボリューム",
        )
            
    # figure
    fig = go.Figure(
        data=trace,
        layout=layout,
        )
    
    return fig


class DashProcessMonitor:

    def start(self):
        self.app.run(debug=True, host='localhost', port=8080)
        webbrowser.open("http://localhost:8080")

    
    def __init__(self, FEMOpt):
        # 引数の処理
        self.FEMOpt = FEMOpt
        
        # application の準備
        self.app = dash.Dash(__name__)

        # layout の設定
        self.app.layout = html.Div(
            html.Div(
                [
                    dcc.Graph(id='hypervolume-graph'),
                    dcc.Graph(id='scatter-matrix-graph'),
                    dcc.Interval(
                        id='interval-component',
                        interval=1*1000, # in milliseconds
                        n_intervals=0
                        )
                    ]
                )
            )

        # callback（分割したほうが並列計算できるらしい）

        # hypervolume
        @self.app.callback(
            Output('hypervolume-graph', 'figure'),
            Input('interval-component', 'n_intervals'))
        def update_hv(_):
            return update_hypervolume(self.FEMOpt)

        # scatter matrix
        @self.app.callback(
            Output('scatter-matrix-graph', 'figure'),
            Input('interval-component', 'n_intervals'))
        def update_sm(_):
            return update_scatter_matrix(self.FEMOpt)

        # フラグが立っていれば更新しない
        @self.app.callback(
            Output('interval-component', 'max_intervals'),
            Input('interval-component', 'n_intervals'))
        def stop_interval(_):
            if self.FEMOpt.should_finish():
                return 0  # Stop the interval
            else:
                return -1  # Allow the interval to run indefinitely

def f(FEMOpt):
    for i in range(10):
        FEMOpt.calc()
    FEMOpt.flag = True

    
if __name__=='__main__':
    class DummyFEMOpt:
        def __init__(self):
            self.flag = False
            self.calc()
            
        def calc(self):
            self.history = pd.DataFrame(
                dict(
                    n_trials=np.arange(10),
                    obj1=np.random.rand(10),
                    obj2=np.random.rand(10),
                    obj3=np.random.rand(10),
                    prm1=np.random.rand(10),
                    prm2=np.random.rand(10),
                    non_domi=np.random.rand(10)>0.5,
                    fit=np.random.rand(10)>0.5,
                    hypervolume=np.random.rand(10),
                    n_trial=np.arange(10),
                    time=[datetime.datetime.now()]*10,
                    )
                )
            sleep(1)
            
        def get_history_columns(self, kind):
            if kind=='parameter':
                return ['prm1', 'prm2']
            if kind=='objective':
                return ['obj1', 'obj2', 'obj3']
        
        def should_finish(self):
            return self.flag

    FEMOpt = DummyFEMOpt()
    
    t = Thread(target=f, args=(FEMOpt,))
    t.start()

    dpm = DashProcessMonitor(FEMOpt)
    dpm.start()
    
    t.join()
    


import dash
from dash import dcc, html
from dash.dependencies import Output, Input
import plotly.graph_objs as go
import plotly.express as px

import numpy as np
import pandas as pd


class DashProcessMonitor:

    def __init__(self, FEMOpt):
        # 引数の処理
        self.FEMOpt = FEMOpt
        
        # application の準備
        self.app = dash.Dash(__name__)
        graph1 = dcc.Graph(id='scatter-graph', animate=False)
        graph2 = dcc.Graph(id='pairplot-graph', animate=False)
        interval_component = dcc.Interval(
            id='interval',
            interval=2*1000,
            n_intervals=0
            )
        self.app.layout = html.Div([graph1, graph2, interval_component])

        # アップデート関数の設定
        # pair-plot の場合、これが一度でも走るとダメなので
        # FEMOpt の init_history が終わったらこれを実行したい。
        # つまり、これの定義を行う関数を別に設定すべきである。
        @self.app.callback(
            Output('pairplot-graph', 'figure'),
            Input('interval', 'n_intervals')
            )
        def update_pp(_):
            return self.update_pp()

        @self.app.callback(
            Output('scatter-graph', 'figure'),
            Input('interval', 'n_intervals')
            )
        def update_hv(_):
            return self.update_hv()


    
    def start(self):
        self.app.run_server(debug=True, host='localhost', port=8050)
        import webbrowser
        webbrowser.open("http://localhost:8050")

    def update_hv(self):
        df = self.FEMOpt.history
        fig = go.Figure()
        x = df['n_trial']
        y = df['hypervolume']
        fig.add_trace(go.Scatter(x=x, y=y, mode='lines'))
        fig.update_layout(
            xaxis_range=[0, x.max()*1.1],
            yaxis_range=[0, y.max()*1.1],
            )
        return fig

        
    def update_pp(self):
        # https://plotly.com/python/splom/
        df = self.FEMOpt.history

        obejctive_names = self.FEMOpt.get_history_columns('objective')
        parameter_names = self.FEMOpt.get_history_columns('parameter')
        index_vals = df['non_domi'].astype('category').cat.codes

        # fig = go.Figure(
        #     data=go.Splom(
        #         dimensions=[dict(label=column, values=df[column]) for column in obejctive_names],
        #         showupperhalf=False,
        #         diagonal_visible=False,
        #         text=df[parameter_names],
        #         marker=dict(
        #             color=index_vals,
        #             showscale=False, # colors encode categorical variables
        #             line_color='white',
        #             line_width=0.5
        #             )
        #         )
        #     )
        # fig.update_layout(
        #     title='pp',
        #     # width=600,
        #     # height=600,
        # )

        fig = px.scatter_matrix(
            df,
            dimensions=obejctive_names,
            color="non_domi",
            symbol="non_domi",
            title="Scatter matrix",
            )
        fig.update_traces(
            diagonal_visible=False,
            showupperhalf=False,
            )
        return fig
    
if __name__=='__main__':
    class DummyFEMOpt:
        def __init__(self):
            self.history = pd.DataFrame(
                dict(
                    n_trials=np.arange(10),
                    obj1=np.random.rand(10),
                    obj2=np.random.rand(10),
                    obj3=np.random.rand(10),
                    prm1=np.random.rand(10),
                    prm2=np.random.rand(10),
                    non_domi=np.random.rand(10)>0.5,
                    )
                )
        def get_history_columns(self, kind):
            if kind=='parameter':
                return ['prm1', 'prm2']
            if kind=='objective':
                return ['obj1', 'obj2', 'obj3']
    dpm = DashProcessMonitor(DummyFEMOpt())
    dpm.start()


# for hint
from __future__ import annotations
from typing import Any, Callable
from abc import ABC, abstractmethod
from warnings import warn

# built-in modules
import os
import sys
import time
import datetime
import inspect
from multiprocessing import Process, Value, Queue
import math
import ast
import tempfile
import shutil
import gc

# 3rd-party modules
import numpy as np
import pandas as pd
from optuna._hypervolume import WFG

# for Femtet
import win32com.client
from win32com.client import Dispatch, constants, CDispatch
from pywintypes import com_error
from femtetutils import util, constant
from ..tools.DispatchUtils import Dispatch_Femtet, Dispatch_Femtet_with_new_process, Dispatch_Femtet_with_specific_pid, _get_pid

# for UI
from PySide6.QtWidgets import QApplication
from ._SimplestUI import SimplestDialog


#### Exception for Femtet error
class ModelError(Exception):
    '''FEM でのモデル更新に失敗した際のエラー'''
    pass


class MeshError(Exception):
    '''FEM でのメッシュ生成に失敗した際のエラー'''
    pass


class SolveError(Exception):
    '''FEM でのソルブに失敗した際のエラー'''
    pass


class PostError(Exception):
    '''FEM 解析前後の目的又は拘束の計算に失敗した際のエラー'''
    pass


class FEMCrash(Exception):
    '''FEM プロセスが異常終了した際のエラー'''
    pass


class FemtetAutomationError(Exception):
    '''その他の Femtet 自動化エラー'''
    pass


class UserInterruption(Exception):
    '''ユーザーによる中断'''
    pass


#### FEM クラス

class FEMSystem(ABC):
    '''
    FEM システムのクラス。
    このクラスを継承して具体的な FEM システムのクラスを作成し、
    そのインスタンスを FemtetOptimizationCore の FEM メンバーに割り当てる。
    '''

    def __init__(self, *args, **kwargs):
        self.args = args
        self.kwargs = kwargs


    @abstractmethod
    def update(self, df) -> None:
        '''
        FEM の制御を実装して df に基づいてモデルを更新し
        objectiveValues と constraintValues を更新する
        '''
        pass


class Femtet(FEMSystem):

    def open(self, femprj_path: str, model_name: str or None = None) -> None:
        '''

        Parameters
        ----------
        femprj_path : str
            .femprj ファイルのパス.
        model_name : str or None, optional
            解析モデルの名前. 指定しない場合、プロジェクトを最後に開いたときのモデルが使われます。
            The default is None.

        Returns
        -------
        None
            DESCRIPTION.

        '''

        self.femprj_path = os.path.abspath(femprj_path)
        self.model_name = model_name
        # 開く
        if self.model_name is None:
            result = self.Femtet.LoadProject(
                self.femprj_path,
                True
            )
        else:
            result = self.Femtet.LoadProjectAndAnalysisModel(
                self.femprj_path,
                self.model_name,
                True
            )
        if not result:
            self.Femtet.ShowLastError()

    def _run_with_catch_error(self, function, iferror=False) -> str:
        # Solve returns False is an error ocured in Femtet
        # Gaudi.Mesh returns number of elements or 0 if error
        # for example, _withcatcherror(Femtet.Solve, False), _withcatcherror(Femtet.Gaudi.Mesh, 0)
        result = function()
        if result == iferror:
            try:
                # ShowLastError() raises com_error
                # that contains the error message of Femtet
                self.Femtet.ShowLastError()  # 必ずエラーが起こる
            except com_error as e:
                info: tuple = e.excepinfo
                error_message: str = info[2]  # 例：'E1033 Model : モデルファイルの保存に失敗しました'
                error_code = error_message.split(' ')[0]
                # ただし、エラーの書式が全部" "区切りかは不明
                return error_code
                # 'E1033':モデル保存失敗
                # 'E4019':Femtetが開いてません
                # 例えば、モデルファイルの保存は再試行したらいい。
                # （なんで保存に失敗するかは謎...sleepしたほうがいいのか。）
        else:
            return ''

    def connect_Femtet(self, strategy: str = 'new', pid: int or None = None) -> str:
        """
        Femtet プロセスとの接続を行います。

        Parameters
        ----------
        strategy : str, optional
            'new' or 'catch' or 'auto'. The default is 'new'.
            'new' のとき、新しい Femtet プロセスを起動し、接続します。
            'catch' のとき、既存の Femtet プロセスと接続します。
            ただし、接続できる Femtet が存在しない場合、エラーを送出します。
            この場合、既存 Femtet プロセスのどれに接続されるかは制御できません。
            また、すでに別のPython又はExcelプロセスと接続されている Femtet との接続を行うことはできません。
            'auto' のとき、まず 'catch' を試行し、失敗した際には 'new' で接続します。
        
        pid : int or None, optional
            接続したい既存の Femtet のプロセス ID. The default is None.
            strategy が catch 以外の時は無視されます。

        Raises
        ------
        Exception
            何らかの理由で Femtet との接続に失敗した際に例外を送出します。

        Returns
        -------
        caught_femtet : str.
            接続された Femtet が既存のものならば 'existing',
            新しいものならば 'new' を返します。

        """

        caught_femtet = None

        if strategy == 'new':
            # 新しい Femtet プロセスを立てて繋ぐ
            self.Femtet, _ = Dispatch_Femtet_with_new_process()
            caught_femtet = 'new'

        elif strategy == 'catch':
            # 既存の Femtet を探して Dispatch する。
            if pid is not None:
                self.Femtet, mypid = Dispatch_Femtet(timeout=5)
            else:
                self.Femtet, mypid = Dispatch_Femtet_with_specific_pid(pid)
            if mypid == 0:
                raise FemtetAutomationError('接続できる Femtet のプロセスが見つかりませんでした。')
            caught_femtet = 'existing'

        elif strategy == 'auto':
            # 既存の Femtet を探して Dispatch する。
            self.Femtet, mypid = Dispatch_Femtet(timeout=5)
            caught_femtet = 'existing'
            if mypid == 0:
                # 失敗したら新しい Femtet プロセスを立てて繋ぐ
                self.Femtet, _ = Dispatch_Femtet_with_new_process()
                caught_femtet = 'new'

        else:
            raise Exception(f'不明な接続方法です：{strategy}')

        # makepy しているかどうかの確認
        if not hasattr(constants, 'STATIC_C'):
            cmd = f'{sys.executable} -m win32com.client.makepy FemtetMacro'
            os.system(cmd)
            print('Python マクロで Femtet 定数を用いる設定がされていませんでした. 設定は自動実行されました.')

        return caught_femtet

    def update_model(self, df: pd.DataFrame) -> None:
        '''
        Femtet の解析モデルを df に従って再構築します。

        Parameters
        ----------
        df : pd.DataFrame
            name, value のカラムを持つこと.

        Raises
        ------
        ModelError
            変数の更新またはモデル再構築に失敗した場合に送出されます。

        Returns
        -------
        None

        '''

        # 変数更新のための処理
        time.sleep(0.1)  # Gaudi がおかしくなる時がある対策
        Gaudi = self.Femtet.Gaudi
        Gaudi.Activate()

        # Femtet の設計変数の更新
        for i, row in df.iterrows():
            name = row['name']
            value = row['value']
            if not (self.Femtet.UpdateVariable(name, value)):
                raise ModelError('変数の更新に失敗しました：変数{name}, 値{value}')

        # 設計変数に従ってモデルを再構築
        # 失敗した場合、変数の更新が反映されない
        if not Gaudi.ReExecute():
            raise ModelError('モデル再構築に失敗しました')

        # インポートを含むモデルに対し ReExecute すると
        # 結果画面のボディツリーでボディが増える対策
        self.Femtet.Redraw()

    def run(self) -> None:
        '''
        Femtet の解析を実行します。

        Raises
        ------
        MeshError
        SolveError

        Returns
        -------
        None.

        '''

        # メッシュを切る
        error_code = self._run_with_catch_error(self.Femtet.Gaudi.Mesh)
        if error_code != '':
            raise MeshError('メッシュ生成に失敗しました。')

        # ソルブする
        error_code = self._run_with_catch_error(self.Femtet.Solve)
        if error_code != '':
            raise SolveError('ソルブに失敗しました。')

        # 次に呼ばれるはずのユーザー定義コスト関数の記述を簡単にするため先に解析結果を開いておく
        if not (self.Femtet.OpenCurrentResult(True)):
            self.Femtet.ShowLastError

    def update(self, df: pd.DataFrame) -> None:
        '''
        update_model と run を実行します。

        Parameters
        ----------
        df : pd.DataFrame
            name, value のカラムを持つこと.

        '''
        self.update_model(df)
        self.run()

    def quit(self):
        # 上書き保存
        self.Femtet.SaveProject(self.Femtet.Project, True)
        # 閉じる
        hwnd = self.Femtet.hWnd
        util.close_femtet(hwnd, 10, True)


class NoFEM(FEMSystem):
    '''
    デバッグ用クラス。
    '''

    def __init__(self, *args, **kwargs):
        self.error_regions = []
        super().__init__(*args, **kwargs)

    def set_error_region(self, f):
        '''
        モデルエラーを起こすための関数 f[[np.ndarray], float] をセットする。
        '''
        self.error_regions.append(f)

    def update(self, df):
        '''
        自身にセットされたerror_regionに抵触する場合、ModelErrorを送出する。
        '''
        for f in self.error_regions:
            if f(df['value'].values) < 0:
                raise ModelError


#### 周辺クラス、関数

def _is_access_Gaudi(f):
    '''
    ユーザー定義関数が Gaudi にアクセスしているかどうかを簡易的に判断する。
    solve前に評価するstrict_constraintでsolve結果にアクセスする違反を
    検出するためのfool-proof策。
    '''
    # 関数fのソースコードを取得
    source = inspect.getsource(f)

    # ソースコードを抽象構文木（AST）に変換
    tree = ast.parse(source)

    # 関数定義を見つける
    for node in ast.walk(tree):
        if isinstance(node, ast.FunctionDef):
            # 関数の第一引数の名前を取得
            first_arg_name = node.args.args[0].arg

            # 関数内の全ての属性アクセスをチェック
            for sub_node in ast.walk(node):
                if isinstance(sub_node, ast.Attribute):
                    # 第一引数に対して 'Gogh' へのアクセスがあるかチェック
                    if (
                            isinstance(sub_node.value, ast.Name)
                            and sub_node.value.id == first_arg_name
                            and sub_node.attr == 'Gogh'
                    ):
                        return True
            # ここまできてもなければアクセスしてない
            return False


def symlog(x):
    '''
    定義域を負領域に拡張したlog関数です。
    多目的最適化における目的関数同士のスケール差により
    意図しない傾向が生ずることのの軽減策として
    内部でsymlog処理を行います。
    '''
    if isinstance(x, np.ndarray):
        ret = np.zeros(x.shape)
        idx = np.where(x >= 0)
        ret[idx] = np.log10(x[idx] + 1)
        idx = np.where(x < 0)
        ret[idx] = -np.log10(1 - x[idx])
    else:
        if x >= 0:
            ret = np.log10(x + 1)
        else:
            ret = -np.log10(1 - x)

    return ret


class Objective:
    prefixForDefault = 'objective'

    def __init__(self, f, args, kwargs, name, direction, FEMOpt):

        self._checkDirection(direction)

        self.f = f
        self.args = args
        self.kwargs = kwargs

        self.name = name
        self.direction = direction

        self.FEMOpt = FEMOpt

    def _checkDirection(self, direction):
        if type(direction) == float or type(direction) == int:
            pass
        elif direction == 'minimize':
            pass
        elif direction == 'maximize':
            pass
        else:
            raise Exception(f'invalid direction "{direction}" for Objective')

    def _convert(self, value: float):
        '''目的関数の値を direction を加味した内部値に変換します。'''
        ret = value
        if type(self.direction) == float or type(self.direction) == int:
            ret = abs(value - self.direction)
        elif self.direction == 'minimize':
            ret = value
        elif self.direction == 'maximize':
            ret = -value

        ret = symlog(ret)

        return ret

    def calc(self):
        # Femtet を継承したクラスの場合第一引数は Femtet であるとする
        if issubclass(self.FEMOpt.FEMClass, Femtet):
            return self.f(*(self.FEMOpt.FEM.Femtet, *self.args), **self.kwargs)
        else:
            return self.f(*self.args, **self.kwargs)


class Constraint:
    prefixForDefault = 'constraint'

    def __init__(self, f, args, kwargs, name, lb, ub, strict, FEMOpt):
        self.f = f
        self.args = args
        self.kwargs = kwargs

        self.name = name
        self.lb = lb
        self.ub = ub
        self.strict = strict

        self.FEMOpt = FEMOpt

    def calc(self):
        # Femtet を継承したクラスの場合第一引数は Femtet であるとする
        if issubclass(self.FEMOpt.FEMClass, Femtet):
            return self.f(*(self.FEMOpt.FEM.Femtet, *self.args), **self.kwargs)
        else:
            return self.f(*self.args, **self.kwargs)


#### 主要抽象クラス

class FemtetOptimizationCore(ABC):

    def __init__(
            self,
            femprj_path: str or None = None,  # None なら既存 Femtet を捕まえる
            model_name: str or None = None,  # None なら開いたときのやつと見做す
            history_path: str = None,  # None なら自動で日付の csv と db を作る
            associated_cad_path: str or None = None,  # None なら pure Femtet と見做す
            FEM: FEMSystem or None = None,  # None なら Femtet or CAD_Femtet と見做す, そうでないならこれを最優先する
    ):

        #### 引数の処理
        if femprj_path is not None:
            self.femprj_path = os.path.abspath(femprj_path)
        else:
            self.femprj_path = femprj_path

        self.model_name = model_name
        self.history_path = history_path
        self.associated_cad_path = associated_cad_path
        self.FEM = FEM


        #### FEM クラスのセットアップ

        self._FEM_args = tuple()
        self._FEM_kwargs = dict()

        # None なら Femtet 系統の処理に入る
        if self.FEM is None:

            # associated_cad_path がなければ pure Femtet にする
            if self.associated_cad_path is None:
                self.FEM = Femtet(*self._FEM_args)

            # associated_cad_path があればそれに応じたインスタンスを作る
            elif self.associated_cad_path.endswith('.prt'):
                from _NX_Femtet import NX_Femtet
                self._FEM_args = (self.associated_cad_path,)
                self.FEM = NX_Femtet(*self._FEM_args)

            elif self.associated_cad_path.endswith('.sldprt'):
                from _SW_Femtet import SW_Femtet
                self._FEM_args = (self.associated_cad_path,)
                self.FEM = SW_Femtet(*self._FEM_args)

            else:
                raise Exception('対応している CAD 拡張子は .prt, .sldprt のみです.')

        # FEM が指定されていればその引数だけ保管しておく（サブプロセスで使う）
        else:
            self._FEM_args = self.FEM.args
            self._FEM_kwargs = self.FEM.kwargs

        # FEM がセットアップできているはずなので FEMClass を覚えておく
        self.FEMClass = type(self.FEM)


        #### Femtet 特有の処理

        # Femtet との接続、および main 実行前に Femtet を解放するかどうかを決める
        self._release_FEM_on_main = False
        if isinstance(self.FEM, Femtet):

            # 既存の Femtet との接続を試み、存在しなければ新規 Femtet を立ち上げる
            caught_femtet = self.FEM.connect_Femtet('auto')

            # 新規 Femtet を立ち上げた場合
            if caught_femtet == 'new':

                # 新しい Femtet を開いたのに femprj が指定されていない場合
                if femprj_path is None:
                    print('-----')
                    print('.femprj ファイルのパスを入力してください。')
                    femprj_path = input('>>> ')
                    self.FEM.open(femprj_path)
                    print('-----')
                    print('Femtet で目的の解析モデルを開き、この画面で Enter を押してください。')
                    input('Enter to continue...')

                # 新しい Femtet で指定された femprj を開く
                else:
                    self.FEM.open(femprj_path, model_name)

                # 新しい Femtet を立ち上げたときのみ, main 前に閉じる
                self._release_FEM_on_main = True

            # 既存 Femtet を開いた場合
            else:

                # femprj などが指定されていない場合、開いているものを正とする
                if self.femprj_path is None:
                    self.femprj_path = os.path.abspath(self.FEM.Femtet.Project)
                    self.model_name = self.FEM.Femtet.AnalysisModelName

                # そうでなければ load_project する
                else:
                    self.FEM.open(self.femprj_path, self.model_name)


                # # 開いている Femtet の prj が指定された femprj と一致しなかった場合、
                # # 上書き load すると開いている Femtet の作業内容が失われる可能性があるので
                # # Exception を送出する
                # else:
                #
                #     # femprj からして違う場合
                #     if self.femprj_path != self.FEM.Femtet.Project:
                #         raise Exception(
                #             '開いている Femtet に接続しましたが、プログラムで指定されているプロジェクトとは違うプロジェクトが開かれています。正しいプロジェクトを開き、Pythonプロセスを再起動してください。PythonからはどのFemtetプロセスに接続するかを制御できないため、目的のFemtetに確実に接続するためには、目的のプロジェクトを開いている以外のFemtetプロセスをすべて終了し、そのプロセスが他のマクロプロセスに接続されていない状態にしてください。')
                #
                #
                #     # femprj は同じだが model が違いうる
                #     else:
                #         # model が指定されていて、かつ違う場合
                #         if self.model_name is not None:
                #             if self.FEM.Femtet.AnalysisModelName != self.model_name:
                #                raise Exception('開いている Femtet に接続しましたが、プログラムで指定されている解析モデルとは違うモデルが開かれています。正しい解析モデルを開き、Pythonプロセスを再起動してください。')


        #### メンバーの宣言
        # ヒストリパスの設定
        if self.history_path is None:
            name = datetime.datetime.now().strftime('%Y_%m_%d_%H_%M')
            self.history_path = os.path.join(os.getcwd(), f'{name}.csv')
        # member
        self.objectives = []
        self.constraints = []
        self.seed = None
        # abstract member
        self._objectives = []
        self._constraints = []
        # per main()
        self.last_execution_time = -1
        self.process_monitor = None
        self.shared_interruption_flag = Value('i', 0)  # 0 or 1, FEMOpt のインスタンスをサブプロセスに渡しても共有されるフラグ
        self.history = None  # parameter などが入ってからでないと初期化もできないので
        # temporal variables per solve
        self._objective_values = None
        self._constraint_values = None
        self._optimization_message = ''
        self._parameter = pd.DataFrame(
            columns=[
                'name',
                'value',
                'lbound',
                'ubound',
                'memo'
            ],
            dtype=object
        )

    def set_random_seed(self, seed: int or None):
        self.seed = seed

    def get_random_seed(self):
        return self.seed

    #### マルチプロセスで pickle するときに COM や fig は持てないし持っていても仕方がないので消す
    def __getstate__(self):
        state = self.__dict__.copy()
        del state['FEM']
        del state['process_monitor']
        return state

    def __setstate__(self, state):
        self.__dict__.update(state)

    # 子プロセスで FEM インスタンスを作成することを想定
    # import 元が干渉するため NX_Femtet と SW_Femtet は現状マルチプロセスにできない
    # この問題さえクリアすれば、即ち femprj と x_t を分ければマルチプロセスは可能
    # ただし SW は複数プロセスの COM 制御が不可能なのでウィンドウ単位で制御するとか工夫が必要
    def set_FEM(self, pid=None):
        # pid は Femtet の場合に使う pid

        # FEMSystem インスタンスの作成
        self.FEM = self.FEMClass(*self._FEM_args)

        # Femtet 固有の処理
        if isinstance(self.FEM, Femtet):
            # pid が指定なしなら新しく開く
            if pid is None:
                self.FEM.connect_Femtet('new')
            # pid が指定ありならそれと繋げる
            else:
                self.FEM.connect_Femtet('catch', pid)
            # いずれにせよ特定のモデルを開く
            self.FEM.open(self.femprj_path, self.model_name)

    def release_FEM(self):
        if isinstance(self.FEM, Femtet):
            self.FEM.quit()

    #### pre-processing methods

    def add_objective(
            self,
            fun: Callable[[CDispatch, Any], float],
            name: str or None = None,
            direction: str or float = 'minimize',
            args: tuple or None = None,
            kwargs: dict or None = None
    ):
        # 引数の処理
        if args is None:
            args = tuple()
        elif type(args) != tuple:
            args = (args,)
        if kwargs is None:
            kwargs = dict()

        # name が指定されなかった場合に自動生成
        if name is None:
            name = self._getUniqueDefaultName(Objective.prefixForDefault, self.objectives)

        # objective オブジェクトを生成
        new_obj = Objective(fun, args, kwargs, name, direction, self)  # サブプロセス内で引っかからないか？

        # 被っているかどうかを判定し上書き or 追加
        isExisting = False
        for i, existing_obj in enumerate(self.objectives):
            if new_obj.name == existing_obj.name:
                isExisting = True
                warn('すでに登録されている名前が追加されました。上書きされます。')
                self.objectives[i] = new_obj
        if not isExisting:
            self.objectives.append(new_obj)

    def add_constraint(
            self,
            fun: Callable[[CDispatch, Any], float],
            name: str or None = None,
            lower_bound: float or None = None,
            upper_bound: float or None = None,
            strict: bool = True,
            args: tuple or None = None,
            kwargs: dict or None = None,
    ):
        # 引数の処理
        if args is None:
            args = tuple()
        elif type(args) != tuple:
            args = (args,)
        if kwargs is None:
            kwargs = dict()

        # strict constraint の場合、solve 前に評価したいので Gaudi へのアクセスを禁ずる
        if strict:
            if _is_access_Gaudi(fun):
                raise Exception(
                    f'関数{fun.__name__}に Gaudi へのアクセスがあります。拘束の計算において解析結果にアクセスするには、strict = False を指定してください。')

        # name が指定されなかった場合に自動生成
        if name is None:
            name = self._getUniqueDefaultName(Constraint.prefixForDefault, self.constraints)

        # Constraint オブジェクトの生成
        new_cns = Constraint(fun, args, kwargs, name, lower_bound, upper_bound, strict, self)

        # 被っているかどうかを判定し上書き or 追加
        isExisting = False
        for i, existing_cns in enumerate(self.constraints):
            if new_cns.name == existing_cns.name:
                isExisting = True
                warn('すでに登録されている名前が追加されました。上書きされます。')
                self.constraints[i] = new_cns
        if not isExisting:
            self.constraints.append(new_cns)

    def add_parameter(
            self,
            name: str,
            initial_value: float or None = None,
            lower_bound: float or None = None,
            upper_bound: float or None = None,
            memo: str = ''
    ):

        # 変数が Femtet にあるかどうかのチェック
        if type(self.FEM) == Femtet:  # 継承した NX_Femtet とかでは Femtet には変数がない
            variable_names = self.FEM.Femtet.GetVariableNames()
            if name in variable_names:
                femtetValue = self.FEM.Femtet.GetVariableValue(name)
            else:
                raise Exception(
                    f'Femtet プロジェクトに変数{name}が設定されていません。現在の Femtet で登録されいてる変数は {variable_names} です。大文字・小文字の区別の注意してください。')

        if initial_value is None:
            initial_value = femtetValue

        d = {
            'name': name,
            'value': initial_value,
            'lbound': lower_bound,
            'ubound': upper_bound,
            'memo': memo,
        }
        df = pd.DataFrame(d, index=[0], dtype=object)

        if len(self._parameter) == 0:
            newdf = df
        else:
            newdf = pd.concat([self._parameter, df], ignore_index=True)

        if self._isDfValid(newdf):
            self._parameter = newdf
        else:
            raise Exception('パラメータの設定が不正です。')

    def set_parameter(self, df: pd.DataFrame):

        if self._isDfValid(df):
            self._parameter = df
        else:
            raise Exception('パラメータの設定が不正です。')

    def get_parameter(self, format: str = 'dict'):
        """
        現在の変数セットを返します。

        Parameters
        ----------
        format : str, optional
            parameter format.  'df', 'values' or 'dict'.
            The default is 'dict'.

        Raises
        ------
        Exception
            DESCRIPTION.

        Returns
        -------
        TYPE
            DESCRIPTION.

        """
        if format == 'df':
            return self._parameter
        elif format == 'values' or format == 'value':
            return self._parameter.value.values
        elif format == 'dict':
            ret = {}
            for i, row in self._parameter.iterrows():
                ret[row['name']] = row.value
            return ret
        else:
            raise Exception('get_parameter() got invalid format: {format}')

    def _set_process_monitor(self):
        # from .visualization._dash import DashProcessMonitor
        # process_monitor = DashProcessMonitor(self)
        # process_monitor.start()
        if len(self.objectives) == 1:
            from .visualization import SimpleProcessMonitor
            self.process_monitor = SimpleProcessMonitor(self)
        else:
            # from .visualization import UpdatableSuperFigure
            from .visualization import MultiobjectivePairPlot
            # from .visualization import HypervolumeMonitor
            # UpdatableSuperFigure(self, HypervolumeMonitor, MultiobjectivePairPlot)
            self.process_monitor = MultiobjectivePairPlot(self)

    def _isDfValid(self, df):
        # column が指定の属性をすべて有しなければ invalid
        if not all([name in df.columns for name in ['name', 'value', 'ubound', 'lbound']]):
            return False

        # value が float でなければ invalid
        for value in df.value.values:
            try:
                float(value)
            except:
                return False

        # ubound または lbound が float 又は None でなければ invalid
        for value in df.ubound.values:
            if value is None:
                continue
            try:
                float(value)
            except:
                return False

        for value in df.lbound.values:
            if value is None:
                continue
            try:
                float(value)
            except:
                return False

        # value, ubound と lbound の上下関係がおかしければ invalid
        for i, row in df.iterrows():
            v = row['value']
            l = row['lbound']
            u = row['ubound']

            if l is None:
                l = v - 1

            if u is None:
                u = v + 1

            if not (l <= v <= u):
                return False

        return True

    def _getUniqueDefaultName(self, prefix, objects):
        names = [obj.name for obj in objects]
        i = 0
        while True:
            name = f'{prefix}_{i}'
            if not (name in names):
                break
            i += 1
        return name

    #### processing methods

    def f(self, x: np.ndarray, *args, **kwargs) -> [float]:
        '''
        現在のパラメータセットを np.ndarray として受け取り、
        FEM 計算を実行して目的関数の内部値のリストを返します。
        この関数は具象クラスの _main、即ち最適化ライブラリの最適化関数に渡されることを想定しています。
        内部値は、与えられた目的関数が最小化問題となるよう変換された値のsymlogです。

        Parameters
        ----------
        x : np.ndarray
            最適化アルゴリズムによって提案された "次の" パラメータセットです。
            get_parameter('values') の値とは一致しません。

        Raises
        ------
        UserInterruption
            ユーザーによる中断指令があった際に送出されます。
            その時点で実行中の解析がすべて終了した時点で最適化が終了します。

        Returns
        -------
        [float]
            変換された目的関数の内部値です。

        '''

        # 中断指令があったら Exception を起こす
        if self.shared_interruption_flag.value == 1:
            self.release_FEM()
            raise UserInterruption('ユーザーによって最適化が取り消されました。')

        # 渡された x がすでに計算されたものであれば
        # objective も constraint も更新する必要がなく
        # そのまま converted objective を返せばいい
        if not self._isCalculated(x):
            # メッセージの更新（反映は後でされる）
            if 'optimization_message' in list(kwargs.keys()):
                self._optimization_message = kwargs['optimization_message']
            else:
                self._optimization_message = ''

            # update parameters
            self._parameter['value'] = x

            # solve(Exception は当面 Optuna で処理する)
            self.FEM.update(self._parameter)
            self._objective_values = [obj.calc() for obj in self.objectives]
            self._constraint_values = [cns.calc() for cns in self.constraints]
            # self._record()
            row = self._get_current_data()
            if hasattr(self, 'queue'):
                self.queue.put(row)

        return [obj._convert(v) for obj, v in zip(self.objectives, self._objective_values)]

    def _isCalculated(self, x):
        # 提案された x が最後に計算したものと一致していれば True, ただし 1 回目の計算なら False
        #    ひとつでも違う  1回目の計算  期待
        #    True            True         False
        #    False           True         False
        #    True            False        False
        #    False           False        True

        # ひとつでも違う
        condition1 = False
        for _x, _p in zip(x, self._parameter['value'].values):
            condition1 = condition1 or (float(_x) != float(_p))
        # 1回目の計算
        condition2 = len(self.history) == 0
        return not (condition1) and not (condition2)

    @abstractmethod
    def _main(self):
        pass

    def _subprocess_main(self, FEMOpt, myidx, shared_allowing_idx, target_pid):
        '''サブプロセスから呼ばれる、_main 関数の worker'''

        #### サブプロセス作成時に破棄されているはずの COM を含みうる FEM を接続
        # 排他処理なので自分の順番が来るまで待つ
        while True:
            if shared_allowing_idx.value == myidx:
                break
            time.sleep(1)

        # 自分の順番が来たら FEM との接続を行う、これが排他処理
        FEMOpt.set_FEM(target_pid)  # 今のところ Femtet 関係なければ 0 が与えられ、これは無視される

        # 接続が終わったら次のプロセスに許可を与える
        shared_allowing_idx.value = shared_allowing_idx.value + 1

        # femprj の干渉を避けるための前処理
        if issubclass(FEMOpt.FEMClass, Femtet):
            # 一時ディレクトリを作成
            td = tempfile.mkdtemp()
            # 一時ファイルに現在のプロジェクトを保存
            # prjname = FEMOpt.FEM.Femtet.ProjectTitle
            prjname = f"subprocess{myidx}"
            tmp_femprj_path = os.path.join(td, f'{prjname}.femprj')
            result = FEMOpt.FEM.Femtet.SaveProject(tmp_femprj_path, True)
            if result == False:
                FEMOpt.FEM.Femtet.ShowLastError()  # raise com_error

        # 最適化の実行
        FEMOpt._main(myidx)

        # FEM の終了
        FEMOpt.release_FEM()

        # 一時フォルダの削除
        if issubclass(FEMOpt.FEMClass, Femtet):
            shutil.rmtree(td)

    def should_finish(self):
        # TODO:もし終了後呼ばれたらFalseを返すので state 設計を何とかする
        if hasattr(self, 'processes') and hasattr(self, 'queue'):
            return self.queue.empty() and all([not p.is_alive() for p in self.processes])
        else:
            if len(self.history) == 0:
                return False
            else:
                return True

    def update_history(self):
        # history の update
        if hasattr(self, 'queue'):
            while True:
                # キューに何かあればサブプロセスからのデータを受け取る
                if not self.queue.empty():
                    row = self.queue.get()
                    self._append_history(row)
                # そうでなければ終了する
                else:
                    break
                time.sleep(0.1)

    def main(self, n_parallel=1, accept_console_control=True):
        '''
        最適化を実行する。

        Parameters
        ----------
        history_path : str or None, optional
            実行結果を保存する csv パス.
        n_parallel : int
            Femtet の並列実行数。1以上の値を指定すること。

        Raises
        ------
        Exception
            add_objective または add_parameter などが呼ばれていない場合に例外を送出します.

        Returns
        -------
        None

        '''
        # 引数の処理
        self.n_parallel = n_parallel

        #### 最適化プロセスの実行前チェック
        # 変数のチェック
        if ((len(self.objectives) == 0)
                or (len(self._parameter) == 0)):
            raise Exception('パラメータまたは目的関数が設定されていません。')

        # ヒストリの初期化
        self._init_history()

        # 中断指令の初期化
        self.shared_interruption_flag.value = 0

        #### プロセスモニタの初期化（ヒストリの初期化が終わってから）
        self._set_process_monitor()

        # 変数チェックのためだけの FEM がある場合、上書き保存して落としておく
        if self._release_FEM_on_main:
            self.release_FEM()

        #### 最適化の開始; サブプロセスを立て、自身は中断待ちを行う。

        # 先に必要分の Femtet を起動しておく
        pids = []
        for i in range(self.n_parallel):
            if issubclass(self.FEMClass, Femtet):
                util.execute_femtet()
                pid = util.get_last_executed_femtet_process_id()
            else:
                pid = 0
            pids.append(pid)

        # サブプロセスの開始
        start = time.time()
        processes = []
        shared_current_allowing_subprocess_id = Value('i', 0)  # Femtet との接続は仕様上排他制御でないとダメ
        self.queue = Queue()  # 行データ受け取りのためのキュー
        for subprocess_id, pid in enumerate(pids):
            p = Process(
                target=self._subprocess_main,
                args=(
                    self,
                    subprocess_id,
                    shared_current_allowing_subprocess_id,
                    pid,
                )
            )
            p.start()
            print(f'subprocess {subprocess_id} start')
            processes.append(p)
        self.processes = processes

        #### history の update 及び ユーザーの中断指令待ち

        # application の作成
        app = QApplication.instance()
        if app == None:
            app = QApplication([])

        dialog = SimplestDialog(
            self,
            self.shared_interruption_flag,
            get_close_flag=self.should_finish,
            fun_to_update=[
                self.update_history,
                self.process_monitor.update,
            ],
        )
        dialog.show()
        app.exec()

        # 全てのサブプロセスの終了を待つ
        for p in processes:
            p.join()

        # 一応
        for p in processes:
            del p
        del processes
        del self.queue
        gc.collect()

        if self.shared_interruption_flag.value == 0:
            print('最適化プロセスは終了しました。')
        else:
            print('最適化プロセスは中断されました。')

        end = time.time()
        self.last_execution_time = end - start  # 秒

        input(f'''計算時間は{self.last_execution_time}秒でした。
結果を確認するには {self.history_path} を開いてください。
終了するには Enter を押してください。''')

    # post-processing methods

    def _genarateHistoryColumnNames(self):
        paramNames = [p for p in self._parameter['name'].values]
        objNames = [obj.name for obj in self.objectives]
        _consNames = [cns.name for cns in self.constraints]
        consNames = []
        for n in _consNames:
            consNames.extend([f'{n}_lower_bound', n, f'{n}_upper_bound', f'{n}_fit'])
        return paramNames, objNames, consNames

    def get_history_columns(self, kind: str = 'objective') -> [str]:
        '''
        get history column names. 

        Parameters
        ----------
        kind : str, optional
            valid kind is: "objective", "parameter", "constraint". The default is 'objective'.

        Returns
        -------
        [str]
            list of column names.

        '''
        p, o, c = self._genarateHistoryColumnNames()
        if kind == 'parameter':
            return p
        elif kind == 'objective':
            return o
        elif kind == 'constraint':
            if len(c) == 0:
                return []
            else:
                return c[1::4]
        else:
            raise Exception(f'invalid kind: {kind}; kind must be "parameter", "objective" or "constraint"')

    def _init_history(self):
        columns = []
        columns.append('n_trial')
        pNames, oNames, cNames = self._genarateHistoryColumnNames()
        columns.extend(pNames)
        columns.extend(oNames)
        columns.extend(cNames)  # 上下限の情報も含む
        columns.append('non_domi')
        columns.append('fit')
        columns.append('hypervolume')
        columns.append('optimization_message')
        columns.append('time')
        self.history = pd.DataFrame(columns=columns)

    def _get_current_data(self):
        # history に保存する値について
        # 現在の値に基づいて集計できるところは集計する
        row = []
        paramValues = list(self._parameter['value'].copy().values)
        objValues = self._objective_values
        consValues = self._constraint_values
        consLbList = [cns.lb for cns in self.constraints]
        consUbList = [cns.ub for cns in self.constraints]
        consData = []
        for v, lb, ub in zip(consValues, consLbList, consUbList):
            consData.extend([lb, v, ub])
            lb = -np.inf if lb is None else lb
            ub = np.inf if ub is None else ub
            fit = (lb <= v <= ub)
            consData.extend([fit])
        row.append(-1)  # temporal data
        row.extend(paramValues)
        row.extend(objValues)
        row.extend(consData)
        row.append(False)  # non_domi, temporal data
        row.append(False)  # fit, temporal data
        row.append(np.nan)  # hypervolume, temporal data
        row.append(self._optimization_message)
        row.append(datetime.datetime.now())  # time
        return row

    def _append_history(self, row):
        # list を series に変換
        pdf = pd.Series(row, index=self.history.columns)

        #### 履歴に依存しない項目の計算
        # 何番目か
        pdf['n_trial'] = len(self.history) + 1

        # 拘束を満たすかどうか
        fit = self._calcConstraintFit(pdf)
        pdf['fit'] = fit

        # series を history に追加
        self.history.loc[len(self.history)] = pdf

        #### 履歴に依存する項目の計算

        # update non-domi(after addition to history)
        self._calcNonDomi()

        # calc hypervolume(after addition to history)
        self._calc_hypervolume()

        # 保存
        try:
            self.history.to_csv(self.history_path, index=None, encoding='shift-jis')
        except PermissionError:
            # excel で開くなど permission error が出る可能性がある。
            # その場合は単にスキップすれば、次の iteration ですべての履歴が保存される
            pass

    def _calcNonDomi(self):
        '''non-dominant 解を計算する'''

        # 目的関数の履歴を取り出してくる
        _, oNames, _ = self._genarateHistoryColumnNames()
        _objectiveValues = self.history[oNames].copy()

        # 最小化問題の座標空間に変換する
        for objective in self.objectives:
            name = objective.name
            _objectiveValues[name] = _objectiveValues[name].map(objective._convert)

            # 非劣解の計算
        non_domi = []
        for i, row in _objectiveValues.iterrows():
            non_domi.append((row > _objectiveValues).product(axis=1).sum(axis=0) == 0)

        # 非劣解の登録
        self.history['non_domi'] = non_domi

        del _objectiveValues

    def _calcConstraintFit(self, pdf):
        '''constraint fit 解を計算する'''

        # 各拘束に対し fit しているかどうかの値を取り出してくる
        _, _, cNames = self._genarateHistoryColumnNames()
        fit = np.all(pdf[[cName for cName in cNames if '_fit' in cName]])

        return fit

    def _calc_hypervolume(self):
        '''
        hypervolume 履歴を更新する
        ※ reference point が変わるたびに hypervolume を計算しなおす必要がある
        [1]Hisao Ishibuchi et al. "Reference Point Specification in Hypercolume Calculation for Fair Comparison and Efficient Search"
        '''
        #### 前準備
        # パレート集合の抽出
        idx = self.history['non_domi']
        pdf = self.history[idx]
        parate_set = pdf[self.get_history_columns('objective')].values
        n = len(parate_set)  # 集合の要素数
        m = len(parate_set.T)  # 目的変数数
        # 長さが 2 以上でないと計算できない
        if n <= 1:
            return np.nan
        # 最小化問題に convert
        for i, objective in enumerate(self.objectives):
            for j in range(n):
                parate_set[j, i] = objective._convert(parate_set[j, i])
                #### reference point の計算[1]
        # 逆正規化のための範囲計算
        maximum = parate_set.max(axis=0)
        minimum = parate_set.min(axis=0)
        # (H+m-1)C(m-1) <= n <= (m-1)C(H+m) になるような H を探す
        H = 0
        while True:
            left = math.comb(H + m - 1, m - 1)
            right = math.comb(H + m, m - 1)
            if left <= n <= right:
                break
            else:
                H += 1
        # H==0 なら r は最大の値
        if H == 0:
            r = 2
        else:
            # r を計算
            r = 1 + 1. / H
        r = 1.01
        # r を逆正規化
        reference_point = r * (maximum - minimum) + minimum

        #### hv 履歴の計算
        wfg = WFG()
        hvs = []
        for i in range(n):
            hv = wfg.compute(parate_set[:i], reference_point)
            if np.isnan(hv):
                hv = 0
            hvs.append(hv)

        # 計算結果を履歴の一部に割り当て
        df = self.history
        df.loc[idx, 'hypervolume'] = np.array(hvs)

        # dominated の行に対して、上に見ていって
        # 最初に見つけた non-domi 行の hypervolume の値を割り当てます
        for i in range(len(df)):
            if df.loc[i, 'non_domi'] == False:
                try:
                    df.loc[i, 'hypervolume'] = df.loc[:i][df.loc[:i]['non_domi']].iloc[-1]['hypervolume']
                except IndexError:
                    # pass # nan のままにする
                    df.loc[i, 'hypervolume'] = 0



from time import sleep

from PyFemtet.opt import FemtetOptuna
from PyFemtet.opt.core import NoFEM

import numpy as np

def objective_x(FEMOpt):
    sleep(1)
    r, theta, fai = FEMOpt.get_parameter('values')
    return r * np.cos(theta) * np.cos(fai)


def objective_y(FEMOpt):
    r, theta, fai = FEMOpt.get_parameter('values')
    return r * np.cos(theta) * np.sin(fai)

def objective_z(FEMOpt):
    r, theta, fai = FEMOpt.get_parameter('values')
    return r * np.sin(theta)

def constraint_y(FEMOpt):
    y = objective_y(FEMOpt)
    return y
    
def constraint_z(FEMOpt):
    z = objective_z(FEMOpt)
    return z

    
if __name__=='__main__':
    FEMOpt = FemtetOptuna(None, FEMClass=NoFEM)
    FEMOpt.add_parameter('r', .5, 0, 1)
    FEMOpt.add_parameter('theta', np.pi/3, 0, 2*np.pi)
    FEMOpt.add_parameter('fai', np.pi/3, -np.pi/2, np.pi/2)
    FEMOpt.add_objective(objective_x, 'x', args=FEMOpt)
    FEMOpt.add_objective(objective_y, 'y', args=FEMOpt)
    FEMOpt.add_objective(objective_z, 'z', args=FEMOpt)
    FEMOpt.add_constraint(constraint_y, 'y<=0', upper_bound=0, args=FEMOpt)
    FEMOpt.add_constraint(constraint_z, 'z<=0', upper_bound=0, args=FEMOpt, strict=False)
    # print(FEMOpt.FEM)
    # print(FEMOpt.FEMClass)
    # print(FEMOpt.constraints)
    # print(FEMOpt.objectives)
    # print(FEMOpt.get_parameter())
    # print(FEMOpt.history)
    # print(FEMOpt.history_path)
    # print(FEMOpt.last_execution_time)
    # FEMOpt._init_history()
    # print(len(FEMOpt.history))
    # print(FEMOpt.f(FEMOpt.get_parameter('values')))
    # print(FEMOpt._objective_values)
    # print(FEMOpt._constraint_values)
    # row = FEMOpt._get_current_data()
    # FEMOpt._append_history(row)
    # FEMOpt._calc_hypervolume()
    # FEMOpt.history['hypervolume']
    

    FEMOpt.main(n_trials=1000, n_parallel=3)
    
    print(len(FEMOpt.history)) # n_trials から prune を抜いた数か

    import optuna
    study = optuna.load_study(study_name=FEMOpt.study_name, storage=FEMOpt.storage_name)
    df = study.trials_dataframe()
    idx = df['state']=='COMPLETE'
    print(len(df[idx]))
    
    print(len(df[idx])==len(FEMOpt.history)) # 中身も見たが OK
    
    print(len(df))
    
    
    # ['number', 'values_0', 'values_1', 'datetime_start', 
    # 'datetime_complete',
    # 'duration', 'params_r', 'params_theta', 'user_attrs_constraint',
    # 'user_attrs_memo', 'system_attrs_constraints',
    # 'system_attrs_fixed_params', 'state']

    
    
    





from PyFemtet.opt import core

from PyFemtet.opt import visualization
from PyFemtet.opt.visualization import _visualizationNameSpace

from PyFemtet.opt.visualization import SimpleProcessMonitor


visualization.__dir__()


